from __future__ import annotations

from typing import TYPE_CHECKING, Literal

import torch
from torch import nn
from torch.nn import functional as F
from torch.nn.common_types import _size_1_t, _size_2_t, _size_3_t
from torch.nn.modules.utils import _pair, _reverse_repeat_tuple, _single, _triple

from .. import params
from .module import BNNModule

if TYPE_CHECKING:
    from jaxtyping import Float
    from torch import Tensor


class _ConvNd(BNNModule):

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: tuple[int, ...],
        stride: tuple[int, ...],
        padding: str | tuple[int, ...],
        dilation: tuple[int, ...],
        transposed: bool,
        output_padding: tuple[int, ...],
        groups: int,
        bias: bool,
        padding_mode: str,
        layer_type: Literal["input", "hidden", "output"] = "hidden",
        cov: params.FactorizedCovariance | None = None,
        parametrization: params.Parametrization = params.MaximalUpdate(),
        device: torch.device = None,
        dtype: torch.dtype = None,
    ) -> None:

        factory_kwargs = {"device": device, "dtype": dtype}
        super().__init__(parametrization=parametrization)

        # Check arguments
        if groups <= 0:
            raise ValueError("groups must be a positive integer.")
        if in_channels % groups != 0:
            raise ValueError("in_channels must be divisible by groups.")
        if out_channels % groups != 0:
            raise ValueError("out_channels must be divisible by groups.")

        valid_padding_strings = {"same", "valid"}
        if isinstance(padding, str):
            if padding not in valid_padding_strings:
                raise ValueError(
                    f"Invalid padding string {padding!r}, should be one of {valid_padding_strings}."
                )
            if padding == "same" and any(s != 1 for s in stride):
                raise ValueError(
                    "padding='same' is not supported for strided convolutions."
                )

        valid_padding_modes = {"zeros", "reflect", "replicate", "circular"}
        if padding_mode not in valid_padding_modes:
            raise ValueError(
                f"padding_mode must be one of {valid_padding_modes}, but got padding_mode='{padding_mode}'."
            )

        # Attributes
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.kernel_size = kernel_size
        self.stride = stride
        self.padding = padding
        self.dilation = dilation
        self.transposed = transposed
        self.output_padding = output_padding
        self.groups = groups
        self.padding_mode = padding_mode
        # `_reversed_padding_repeated_twice` is the padding to be passed to
        # `F.pad` if needed (e.g., for non-zero padding types that are
        # implemented as two ops: padding + conv). `F.pad` accepts paddings in
        # reverse order than the dimension.
        if isinstance(self.padding, str):
            self._reversed_padding_repeated_twice = [0, 0] * len(kernel_size)
            if padding == "same":
                for d, k, i in zip(
                    dilation, kernel_size, range(len(kernel_size) - 1, -1, -1)
                ):
                    total_padding = d * (k - 1)
                    left_pad = total_padding // 2
                    self._reversed_padding_repeated_twice[2 * i] = left_pad
                    self._reversed_padding_repeated_twice[2 * i + 1] = (
                        total_padding - left_pad
                    )
        else:
            self._reversed_padding_repeated_twice = _reverse_repeat_tuple(
                self.padding, 2
            )
        if layer_type not in ["input", "hidden", "output"]:
            raise ValueError(
                f"Invalid layer type '{layer_type}'. Must be one of ['input', 'hidden', 'output']."
            )
        self.layer_type = layer_type

        # Parameters
        weight_shape = (
            (in_channels, out_channels // groups, *kernel_size)
            if transposed
            else (out_channels, in_channels // groups, *kernel_size)
        )
        mean_param_dict = {
            **{"weight": torch.empty(weight_shape, **factory_kwargs)},
            **(
                {"bias": torch.empty(out_channels, **factory_kwargs)}
                if bias
                else {"bias": None}
            ),
        }
        if cov is None:
            self.params = nn.ParameterDict(mean_param_dict)
            self.params.cov = None
        else:
            self.params = params.GaussianParameter(
                mean=mean_param_dict,
                cov=cov,
            )

        self.reset_parameters()

    @property
    def weight(self) -> nn.Parameter:
        return self.params.weight

    @property
    def bias(self) -> nn.Parameter:
        return self.params.bias

    def reset_parameters(
        self,
    ) -> None:
        """Reset the parameters of the module."""
        fan_in, fan_out = nn.init._calculate_fan_in_and_fan_out(self.params.weight)

        # Mean parameters
        mean_parameter_scales = {}
        mean_parameter_scales["weight"] = self.parametrization.weight_init_scale(
            fan_in, fan_out, layer_type=self.layer_type
        )
        nn.init.normal_(self.params.weight, mean=0, std=mean_parameter_scales["weight"])

        if self.params.bias is not None:
            mean_parameter_scales["bias"] = self.parametrization.bias_init_scale(
                fan_in, fan_out, layer_type=self.layer_type
            )
            nn.init.normal_(self.params.bias, mean=0, std=mean_parameter_scales["bias"])

        # Covariance parameters
        if self.params.cov is not None:
            self.params.cov.reset_parameters(mean_parameter_scales)

    def parameters_and_lrs(
        self,
        lr: float,
        optimizer: Literal["SGD", "Adam"] = "SGD",
    ) -> list[dict[str, Tensor | float]]:
        fan_in, fan_out = nn.init._calculate_fan_in_and_fan_out(self.params.weight)

        mean_parameter_lr_scales = {}
        mean_parameter_lr_scales["weight"] = self.parametrization.weight_lr_scale(
            fan_in, fan_out, optimizer=optimizer, layer_type=self.layer_type
        )
        param_groups = [
            {
                "params": self.params.weight,
                "lr": lr * mean_parameter_lr_scales["weight"],
            }
        ]

        if self.params.bias is not None:
            mean_parameter_lr_scales["bias"] = self.parametrization.bias_lr_scale(
                fan_in, fan_out, optimizer=optimizer, layer_type=self.layer_type
            )
            param_groups += [
                {
                    "params": self.params.bias,
                    "lr": lr * mean_parameter_lr_scales["bias"],
                }
            ]

        if self.params.cov is not None:
            for name, param in self.params.cov.named_parameters():
                lr_scaling = 1.0
                if "weight" in name:
                    lr_scaling = mean_parameter_lr_scales["weight"]
                elif "bias" in name:
                    lr_scaling = mean_parameter_lr_scales["bias"]
                param_groups += [
                    {
                        "params": param,
                        "lr": lr * lr_scaling * self.params.cov.lr_scaling[name],
                    }
                ]

        return param_groups

    def extra_repr(self):
        s = (
            "{in_channels}, {out_channels}, kernel_size={kernel_size}"
            ", stride={stride}"
        )
        if self.padding != (0,) * len(self.padding):
            s += ", padding={padding}"
        if self.dilation != (1,) * len(self.dilation):
            s += ", dilation={dilation}"
        if self.output_padding != (0,) * len(self.output_padding):
            s += ", output_padding={output_padding}"
        if self.groups != 1:
            s += ", groups={groups}"
        if self.bias is None:
            s += ", bias=False"
        if self.padding_mode != "zeros":
            s += ", padding_mode={padding_mode}"
        return s.format(**self.__dict__)

    def __setstate__(self, state):
        super().__setstate__(state)
        if not hasattr(self, "padding_mode"):
            self.padding_mode = "zeros"

    def _load_from_state_dict(
        self,
        state_dict: dict,
        prefix: str,
        local_metadata: dict,
        strict: bool,
        missing_keys: list[str],
        unexpected_keys: list[str],
        error_msgs: list[str],
    ):

        # Ensure compatibility with nn.Linear
        if prefix + "weight" in state_dict:
            state_dict[prefix + "params.weight"] = state_dict.pop(prefix + "weight")
        if prefix + "bias" in state_dict:
            state_dict[prefix + "params.bias"] = state_dict.pop(prefix + "bias")

        super()._load_from_state_dict(
            state_dict,
            prefix,
            local_metadata,
            strict,
            missing_keys,
            unexpected_keys,
            error_msgs,
        )

    def _conv_forward(
        self,
        input: Float[Tensor, "*batch in_channel *in_feature"],
        weight: Float[Tensor, "out_channel in_channels_per_group kernel_size"],
        bias: Float[Tensor, "out_channel"] | None,
    ) -> Float[Tensor, "*batch out_channel *out_feature"]:
        raise NotImplementedError

    def forward(
        self,
        input: Float[Tensor, "*sample batch *in_feature"],
        /,
        sample_shape: torch.Size = torch.Size([]),
        generator: torch.Generator | None = None,
        input_contains_samples: bool = False,
        parameter_samples: dict[str, Float[Tensor, "*sample parameter"]] | None = None,
    ) -> Float[Tensor, "*sample *batch out_channel *out_feature"]:

        if not input_contains_samples:
            # Repeat input for each desired sample
            input = input.expand(*sample_shape, *input.shape)

        if parameter_samples is None:
            if self.params.cov is None:
                # If there's no covariance, expand the mean parameters into parameter samples
                parameter_samples = {
                    "weight": self.params.weight.expand(
                        *sample_shape, *self.params.weight.shape
                    ),
                    "bias": (
                        self.params.bias.expand(*sample_shape, *self.params.bias.shape)
                        if self.bias is not None
                        else None
                    ),
                }
            else:
                # Sample one set of parameters per sample
                parameter_samples = self.params.sample(
                    sample_shape=sample_shape, generator=generator
                )

        # Forward pass with sampled parameters
        if torch.Size(sample_shape) == torch.Size([]):
            # For a single sample, call _conv_forward directly
            return self._conv_forward(
                input,
                parameter_samples["weight"],
                (parameter_samples["bias"] if self.bias is not None else None),
            )
        else:
            # In case we have multiple sample dimensions, flatten, vmap and unflatten into sample_shape
            flattened_conv_result = torch.vmap(
                self._conv_forward,
                in_dims=(0, 0, 0 if self.bias is not None else None),
            )(
                input.flatten(start_dim=0, end_dim=len(sample_shape) - 1),
                parameter_samples["weight"].flatten(
                    start_dim=0, end_dim=len(sample_shape) - 1
                ),
                (
                    parameter_samples["bias"].flatten(
                        start_dim=0, end_dim=len(sample_shape) - 1
                    )
                    if self.bias is not None
                    else None
                ),
            )

            return flattened_conv_result.unflatten(0, sample_shape)


class Conv1d(_ConvNd):
    r"""Applies a 1D convolution over an input signal composed of several input planes.

    In the simplest case, the output value of the layer with input size
    :math:`(N, C_{\text{in}}, L)` and output :math:`(N, C_{\text{out}}, L_{\text{out}})` can be
    precisely described as:

    .. math::
        \text{out}(N_i, C_{\text{out}_j}) = \text{bias}(C_{\text{out}_j}) +
        \sum_{k = 0}^{C_{in} - 1} \text{weight}(C_{\text{out}_j}, k)
        \star \text{input}(N_i, k)

    where :math:`\star` is the valid `cross-correlation`_ operator,
    :math:`N` is a batch size, :math:`C` denotes a number of channels,
    :math:`L` is a length of signal sequence.

    :param in_channels:     Number of channels in the input image.
    :param out_channels:    Number of channels produced by the convolution.
    :param kernel_size:     Size of the convolving kernel.
    :param stride:          Stride of the convolution.
    :param padding:         Padding added to both sides of the input.
    :param dilation:        Spacing between kernel elements.
    :param groups:          Number of blocked connections from input channels to output channels.
    :param bias:            If ``True``, adds a learnable bias to the output.
    :param padding_mode:    ``'zeros'``, ``'reflect'``, ``'replicate'`` or ``'circular'``. Default is ``'zeros'``.
    :param layer_type:      Type of the layer. Can be one of "input", "hidden", or "output".
        Controls the initialization and learning rate scaling of the parameters.
    :param cov:             The covariance of the parameters.
    :param parametrization: The parametrization to use. Defines the initialization
        and learning rate scaling for the parameters of the module.
    :param device:          The device on which to place the tensor.
    :param dtype:           The desired data type of the returned tensor.
    """

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: _size_1_t,
        stride: _size_1_t = 1,
        padding: str | _size_1_t = 0,
        dilation: _size_1_t = 1,
        groups: int = 1,
        bias: bool = True,
        padding_mode: str = "zeros",
        layer_type: Literal["input", "hidden", "output"] = "hidden",
        cov: params.FactorizedCovariance | None = None,
        parametrization: params.Parametrization = params.MaximalUpdate(),
        device: torch.device = None,
        dtype: torch.dtype = None,
    ) -> None:
        factory_kwargs = {"device": device, "dtype": dtype}

        kernel_size_ = _single(kernel_size)
        stride_ = _single(stride)
        padding_ = padding if isinstance(padding, str) else _single(padding)
        dilation_ = _single(dilation)
        super().__init__(
            in_channels,
            out_channels,
            kernel_size_,
            stride_,
            padding_,
            dilation_,
            False,
            _single(0),
            groups,
            bias,
            padding_mode,
            layer_type=layer_type,
            cov=cov,
            parametrization=parametrization,
            **factory_kwargs,
        )

    def _conv_forward(
        self,
        input: Float[Tensor, "*batch in_channel in_feature"],
        weight: Float[Tensor, "out_channel in_channels_per_group kernel_size"],
        bias: Float[Tensor, "out_channel"] | None,
    ) -> Float[Tensor, "*batch out_channel out_feature"]:
        if self.padding_mode != "zeros":
            input = F.pad(
                input, self._reversed_padding_repeated_twice, mode=self.padding_mode
            )
            padding = _single(0)
        else:
            padding = self.padding

        return F.conv1d(
            input, weight, bias, self.stride, padding, self.dilation, self.groups
        )


class Conv2d(_ConvNd):
    r"""Applies a 2D convolution over an input signal composed of several input planes.

    In the simplest case, the output value of the layer with input size
    :math:`(N, C_{\text{in}}, H, W)` and output :math:`(N, C_{\text{out}}, H_{\text{out}}, W_{\text{out}})`
    can be precisely described as:

    .. math::
        \text{out}(N_i, C_{\text{out}_j}) = \text{bias}(C_{\text{out}_j}) +
        \sum_{k = 0}^{C_{\text{in}} - 1} \text{weight}(C_{\text{out}_j}, k) \star \text{input}(N_i, k)


    where :math:`\star` is the valid 2D `cross-correlation`_ operator,
    :math:`N` is a batch size, :math:`C` denotes a number of channels,
    :math:`H` is a height of input planes in pixels, and :math:`W` is
    width in pixels.

    :param in_channels:     Number of channels in the input image.
    :param out_channels:    Number of channels produced by the convolution.
    :param kernel_size:     Size of the convolving kernel.
    :param stride:          Stride of the convolution.
    :param padding:         Padding added to all four sides of the input.
    :param dilation:        Spacing between kernel elements.
    :param groups:          Number of blocked connections from input channels to output channels.
    :param bias:            If ``True``, adds a learnable bias to the output.
    :param padding_mode:    ``'zeros'``, ``'reflect'``, ``'replicate'`` or ``'circular'``. Default is ``'zeros'``.
    :param layer_type:      Type of the layer. Can be one of "input", "hidden", or "output".
        Controls the initialization and learning rate scaling of the parameters.
    :param cov:             The covariance of the parameters.
    :param parametrization: The parametrization to use. Defines the initialization
        and learning rate scaling for the parameters of the module.
    :param device:          The device on which to place the tensor.
    :param dtype:           The desired data type of the returned tensor.
    """

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: _size_2_t,
        stride: _size_2_t = 1,
        padding: str | _size_2_t = 0,
        dilation: _size_2_t = 1,
        groups: int = 1,
        bias: bool = True,
        padding_mode: str = "zeros",
        layer_type: Literal["input", "hidden", "output"] = "hidden",
        cov: params.FactorizedCovariance | None = None,
        parametrization: params.Parametrization = params.MaximalUpdate(),
        device: torch.device = None,
        dtype: torch.dtype = None,
    ) -> None:
        factory_kwargs = {"device": device, "dtype": dtype}

        kernel_size_ = _pair(kernel_size)
        stride_ = _pair(stride)
        padding_ = padding if isinstance(padding, str) else _pair(padding)
        dilation_ = _pair(dilation)
        super().__init__(
            in_channels,
            out_channels,
            kernel_size_,
            stride_,
            padding_,
            dilation_,
            False,
            _pair(0),
            groups,
            bias,
            padding_mode,
            layer_type=layer_type,
            cov=cov,
            parametrization=parametrization,
            **factory_kwargs,
        )

    def _conv_forward(
        self,
        input: Float[Tensor, "*batch in_channel in_feature0 in_feature1"],
        weight: Float[Tensor, "out_channel in_channels_per_group kernel_size"],
        bias: Float[Tensor, "out_channel"] | None,
    ) -> Float[Tensor, "*batch out_channel out_feature0 out_feature1"]:
        if self.padding_mode != "zeros":
            input = F.pad(
                input, self._reversed_padding_repeated_twice, mode=self.padding_mode
            )
            padding = _pair(0)
        else:
            padding = self.padding

        return F.conv2d(
            input, weight, bias, self.stride, padding, self.dilation, self.groups
        )


class Conv3d(_ConvNd):
    r"""Applies a 3D convolution over an input signal composed of several input planes.

    In the simplest case, the output value of the layer with input size :math:`(N, C_{in}, D, H, W)`
    and output :math:`(N, C_{out}, D_{out}, H_{out}, W_{out})` can be precisely described as:

    .. math::
        out(N_i, C_{out_j}) = bias(C_{out_j}) +
                                \sum_{k = 0}^{C_{in} - 1} weight(C_{out_j}, k) \star input(N_i, k)

    where :math:`\star` is the valid 3D `cross-correlation`_ operator

    :param in_channels:     Number of channels in the input image.
    :param out_channels:    Number of channels produced by the convolution.
    :param kernel_size:     Size of the convolving kernel.
    :param stride:          Stride of the convolution.
    :param padding:         Padding added to all six sides of the input.
    :param dilation:        Spacing between kernel elements.
    :param groups:          Number of blocked connections from input channels to output channels.
    :param bias:            If ``True``, adds a learnable bias to the output.
    :param padding_mode:    ``'zeros'``, ``'reflect'``, ``'replicate'`` or ``'circular'``. Default is ``'zeros'``.
    :param layer_type:      Type of the layer. Can be one of "input", "hidden", or "output".
        Controls the initialization and learning rate scaling of the parameters.
    :param cov:             The covariance of the parameters.
    :param parametrization: The parametrization to use. Defines the initialization
        and learning rate scaling for the parameters of the module.
    :param device:          The device on which to place the tensor.
    :param dtype:           The desired data type of the returned tensor.
    """

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: _size_3_t,
        stride: _size_3_t = 1,
        padding: str | _size_3_t = 0,
        dilation: _size_3_t = 1,
        groups: int = 1,
        bias: bool = True,
        padding_mode: str = "zeros",
        layer_type: Literal["input", "hidden", "output"] = "hidden",
        cov: params.FactorizedCovariance | None = None,
        parametrization: params.Parametrization = params.MaximalUpdate(),
        device: torch.device = None,
        dtype: torch.dtype = None,
    ) -> None:
        factory_kwargs = {"device": device, "dtype": dtype}
        kernel_size_ = _triple(kernel_size)
        stride_ = _triple(stride)
        padding_ = padding if isinstance(padding, str) else _triple(padding)
        dilation_ = _triple(dilation)
        super().__init__(
            in_channels,
            out_channels,
            kernel_size_,
            stride_,
            padding_,
            dilation_,
            False,
            _triple(0),
            groups,
            bias,
            padding_mode,
            layer_type=layer_type,
            cov=cov,
            parametrization=parametrization,
            **factory_kwargs,
        )

    def _conv_forward(
        self,
        input: Float[Tensor, "*batch in_channel in_feature0 in_feature1"],
        weight: Float[Tensor, "out_channel in_channels_per_group kernel_size"],
        bias: Float[Tensor, "out_channel"] | None,
    ) -> Float[Tensor, "*batch out_channel out_feature0 out_feature1"]:
        if self.padding_mode != "zeros":
            input = F.pad(
                input, self._reversed_padding_repeated_twice, mode=self.padding_mode
            )
            padding = _triple(0)
        else:
            padding = self.padding

        return F.conv3d(
            input, weight, bias, self.stride, padding, self.dilation, self.groups
        )



"""Pre-defined models."""

from torch import nn

from .. import bnn
from .ensemble import Ensemble
from .lenet5 import LeNet5
from .mlp import MLP
from .resnet import (
    ResNet,
    ResNet18,
    ResNet34,
    ResNet50,
    ResNet101,
    ResNeXt50_32X4D,
    ResNeXt101_32X8D,
    ResNeXt101_64X4D,
    WideResNet50,
    WideResNet101,
)

___all__ = [
    "Ensemble",
    "LeNet5",
    "MLP",
    "ResNet",
    "ResNet18",
    "ResNet34",
    "ResNet50",
    "ResNet101",
    "ResNeXt50_32X4D",
    "ResNeXt101_32X8D",
    "ResNeXt101_64X4D",
    "WideResNet50",
    "WideResNet101",
    "as_torch_model",
]



import copy
import math

import numpy as np
import numpy.testing as npt
import torch
from torch import nn

from inferno import bnn
from inferno.bnn import params

import pytest


@pytest.mark.parametrize("batch_shape", [(), (1,), (3,), (4, 1, 2)])
@pytest.mark.parametrize("in_features, out_features", [(5, 3), (3, 2)])
def test_generalizes_pytorch_linear_layer(batch_shape, in_features, out_features):
    """Test whether the BNN linear layer generalizes the PyTorch linear layer."""
    generator = torch.Generator().manual_seed(0)

    input = torch.randn(batch_shape + (in_features,), generator=generator)
    linear = torch.nn.Linear(in_features=in_features, out_features=out_features)
    linear.weight = torch.nn.Parameter(linear.weight)
    linear.bias = torch.nn.Parameter(linear.bias)

    bnn_linear = bnn.Linear(in_features=in_features, out_features=out_features)
    bnn_linear.params.weight = torch.nn.Parameter(linear.weight)
    bnn_linear.params.bias = torch.nn.Parameter(linear.bias)

    npt.assert_allclose(
        linear(input).detach().numpy(),
        bnn_linear(input, generator=generator).detach().numpy(),
        atol=1e-6,
        rtol=1e-6,
    )


@pytest.mark.parametrize(
    "linear_layer_to_load",
    [
        bnn.Linear(3, 2),
        bnn.Linear(3, 2, cov=params.FactorizedCovariance()),
        nn.Linear(3, 2),
    ],
)
def test_load_from_state_dict(linear_layer_to_load):
    """Test whether the load_from_state_dict method is working for torch and inferno
    linear layers."""
    state_dict = linear_layer_to_load.state_dict()
    new_linear_layer = bnn.Linear(
        linear_layer_to_load.in_features,
        linear_layer_to_load.out_features,
        cov=params.FactorizedCovariance(),
    )
    new_linear_layer.load_state_dict(
        state_dict,
        strict=hasattr(linear_layer_to_load, "params.cov"),
    )

    prefix = "params." if isinstance(linear_layer_to_load, bnn.BNNModule) else ""

    npt.assert_allclose(
        new_linear_layer.params.weight.detach().numpy(),
        state_dict[prefix + "weight"].detach().numpy(),
    )
    npt.assert_allclose(
        new_linear_layer.params.bias.detach().numpy(),
        state_dict[prefix + "bias"].detach().numpy(),
    )

    if hasattr(linear_layer_to_load, "params.cov"):
        npt.assert_allclose(
            new_linear_layer.params.cov_params.detach().numpy(),
            state_dict[prefix + "cov_params"].detach().numpy(),
        )


@pytest.mark.parametrize("seed", [0, 45234, 42])
def test_forward_is_deterministic_given_generator(seed):
    """Test whether the forward method is deterministic given a generator."""
    linear_layer = bnn.Linear(5, 3, cov=params.FactorizedCovariance())

    input = torch.randn(3, 5, generator=torch.Generator().manual_seed(seed + 2452345))
    output1 = linear_layer(input, generator=torch.Generator().manual_seed(seed))
    output2 = linear_layer(input, generator=torch.Generator().manual_seed(seed))

    npt.assert_allclose(output1.detach().numpy(), output2.detach().numpy())


@pytest.mark.parametrize("sample_shape", [(), (1,), (2,), (3, 2)])
@pytest.mark.parametrize("batch_shape", [(), (1,), (3,), (4, 1, 2)])
@pytest.mark.parametrize(
    "linear_layer",
    [
        bnn.Linear(5, 3, cov=params.FactorizedCovariance()),
        bnn.Linear(3, 2, bias=False, cov=params.LowRankCovariance(2)),
        bnn.Linear(3, 2, bias=False, cov=None),
    ],
)
def test_shape(sample_shape, batch_shape, linear_layer):
    """Test whether the output shape is correct."""
    generator = torch.Generator().manual_seed(0)
    input = torch.randn(*batch_shape, linear_layer.in_features, generator=generator)
    output = linear_layer(input, sample_shape=sample_shape, generator=generator)
    assert output.shape == (*sample_shape, *batch_shape, linear_layer.out_features)


def test_register_forward_hook():
    """Test whether the register_forward_hook method is working."""
    generator = torch.Generator().manual_seed(0)

    linear_layer = bnn.Linear(5, 3, cov=params.FactorizedCovariance())
    test_dict = {"hook_has_fired": False}

    def hook(module, input, output):
        test_dict["hook_has_fired"] = True

    linear_layer.register_forward_hook(hook)
    linear_layer(torch.randn(3, 5), generator=generator)

    assert test_dict["hook_has_fired"]


def random_fourier_features(
    x,
    num_features=100,
    outputscale=1.0,
    lengthscale=0.2,
    generator: torch.Generator | None = None,
):

    # Generate frequencies and phases
    frequencies = (
        torch.randn(x.shape[1], num_features, generator=generator) / lengthscale
    )
    phases = 2 * np.pi * torch.rand(num_features, generator=generator)

    # Compute random Fourier features
    projected = x @ frequencies + phases
    features = outputscale / np.sqrt(num_features) * torch.cos(projected)
    return features


@pytest.mark.parametrize(
    "momentum,nesterov",
    [
        (0.0, False),
        (0.9, False),
        (0.9, True),
    ],
)
@pytest.mark.parametrize(
    "cov",
    [params.LowRankCovariance(5), params.FactorizedCovariance()],
    ids=lambda x: x.__class__.__name__,
)
def test_linear_regression_nullspace_invariance(momentum, nesterov, cov):
    """
    Tests whether Bayesian linear regression with a factorized covariance leaves the prior
    invariant in the nullspace of the data.
    """
    torch.manual_seed(2452345)

    with torch.no_grad():
        # Data

        def feature_map(x):
            return random_fourier_features(x, num_features=100)

        def f(x):
            return torch.sin(x * 2 * np.pi)

        def y(x, noise_scale=0.3):
            return f(x).squeeze() + noise_scale * torch.randn(x.shape[0])

        num_train_data = 100

        X_train_raw = torch.concatenate(
            [
                2 * (torch.rand((num_train_data, 1)) - 0.5),
            ],
            dim=0,
        )
        X_train = feature_map(X_train_raw)
        y_train = y(X_train_raw)
        U, S, VT = torch.linalg.svd(X_train, full_matrices=True)
        V = VT.mT
        # V_range = V[:, : X_train.shape[0]]
        V_null = V[:, X_train.shape[0] :]

        train_dataset = torch.utils.data.TensorDataset(X_train, y_train)

        # Dataloader
        batch_size = 64
        train_dataloader = torch.utils.data.DataLoader(
            train_dataset,
            batch_size=batch_size,
            shuffle=True,
        )

    # Model
    model = bnn.Linear(X_train.shape[-1], 1, bias=False, cov=cov)
    model.reset_parameters()
    prior_model = copy.deepcopy(model)

    # Loss function
    loss_fn = nn.MSELoss()

    # Optimizer
    optimizer = torch.optim.SGD(
        params=model.parameters(),
        lr=1e-1,
        momentum=momentum,
        nesterov=nesterov,
    )

    # Training loop
    num_epochs = 20
    num_samples_train_loss = 32
    for epoch in range(num_epochs):
        model.train()

        for X_batch, y_batch in iter(train_dataloader):
            optimizer.zero_grad()

            # Prediction
            y_batch_pred = model(
                X_batch, sample_shape=(num_samples_train_loss,)
            ).squeeze(-1)

            # Loss
            loss = loss_fn(y_batch_pred, y_batch.expand(num_samples_train_loss, -1))

            # Optimizer step
            loss.backward()
            optimizer.step()

        with torch.no_grad():
            model.eval()
            prior_cov_params = torch.vstack(
                [
                    factor_param.view(-1, prior_model.params.cov.rank)
                    for factor_param in prior_model.params.cov.factor.values()
                    if factor_param is not None
                ]
            )
            cov_params = torch.vstack(
                [
                    factor_param.view(-1, model.params.cov.rank)
                    for factor_param in model.params.cov.factor.values()
                    if factor_param is not None
                ]
            )

            null_space_proj_mean = V_null.mT @ model.params.weight.ravel()
            null_space_proj_cov = V_null.mT @ cov_params @ cov_params.mT @ V_null

            null_space_proj_prior_mean = V_null.mT @ prior_model.params.weight.ravel()
            null_space_proj_prior_cov = (
                V_null.mT @ prior_cov_params @ prior_cov_params.mT @ V_null
            )

            npt.assert_allclose(
                null_space_proj_mean.numpy(),
                null_space_proj_prior_mean.numpy(),
                atol=1e-5,
                rtol=1e-5,
            )

            npt.assert_allclose(
                null_space_proj_cov.numpy(),
                null_space_proj_prior_cov.numpy(),
                atol=1e-5,
                rtol=1e-5,
            )


@pytest.mark.parametrize(
    "momentum,nesterov",
    [
        (0.0, False),
        (0.9, False),
        (0.9, True),
    ],
)
@pytest.mark.parametrize(
    "cov",
    [params.LowRankCovariance(5), params.FactorizedCovariance()],
    ids=lambda x: x.__class__.__name__,
)
def test_logistic_regression_nullspace_invariance(momentum, nesterov, cov):
    """
    Tests whether Bayesian logistic regression with a factorized covariance leaves the prior
    invariant in the nullspace of the data.
    """
    torch.manual_seed(2452345)

    with torch.no_grad():
        # Data

        def feature_map(x):
            return random_fourier_features(x, num_features=100)

        def y(x):
            return torch.as_tensor(
                (x @ torch.ones(input_dim)) >= 0, dtype=torch.float32
            )

        input_dim = 2
        num_train_data = 40

        X_train_raw = torch.concatenate(
            [
                0.4 * torch.randn((num_train_data // 2, input_dim))
                + torch.ones(input_dim),
                0.4 * torch.randn((num_train_data // 2, input_dim))
                - torch.ones(input_dim),
            ],
            axis=0,
        )
        X_train = feature_map(X_train_raw)
        y_train = y(X_train_raw)
        U, S, VT = torch.linalg.svd(X_train, full_matrices=True)
        V = VT.mT
        V_null = V[:, X_train.shape[0] :]

        train_dataset = torch.utils.data.TensorDataset(X_train, y_train)

        # Dataloader
        batch_size = 64
        train_dataloader = torch.utils.data.DataLoader(
            train_dataset,
            batch_size=batch_size,
            shuffle=True,
        )

    # Model
    model = bnn.Sequential(
        bnn.Linear(X_train.shape[-1], 1, bias=False, cov=cov),
        nn.Flatten(-2, -1),
        nn.Sigmoid(),
        parametrization=params.Standard(),
    )
    model[-3].reset_parameters()
    prior_model = copy.deepcopy(model)

    # Loss function
    loss_fn = nn.BCELoss()

    # Optimizer
    optimizer = torch.optim.SGD(
        params=model.parameters(),
        lr=1e0,
        momentum=momentum,
        nesterov=nesterov,
    )

    # Training loop
    num_epochs = 20
    num_samples_train_loss = 32
    for epoch in range(num_epochs):
        model.train()

        for X_batch, y_batch in iter(train_dataloader):
            optimizer.zero_grad()

            # Prediction
            y_batch_pred = model(X_batch, sample_shape=(num_samples_train_loss,))

            # Loss
            loss = loss_fn(y_batch_pred, y_batch.expand(num_samples_train_loss, -1))

            # Optimizer step
            loss.backward()
            optimizer.step()

        with torch.no_grad():
            model.eval()
            prior_cov_params = torch.vstack(
                [
                    factor_param.view(-1, prior_model[-3].params.cov.rank)
                    for factor_param in prior_model[-3].params.cov.factor.values()
                    if factor_param is not None
                ]
            )
            cov_params = torch.vstack(
                [
                    factor_param.view(-1, model[-3].params.cov.rank)
                    for factor_param in model[-3].params.cov.factor.values()
                    if factor_param is not None
                ]
            )

            null_space_proj_mean = V_null.mT @ model[-3].params.weight.ravel()
            null_space_proj_cov = V_null.mT @ cov_params @ cov_params.mT @ V_null

            null_space_proj_prior_mean = (
                V_null.mT @ prior_model[-3].params.weight.ravel()
            )
            null_space_proj_prior_cov = (
                V_null.mT @ prior_cov_params @ prior_cov_params.mT @ V_null
            )

            npt.assert_allclose(
                null_space_proj_mean.numpy(),
                null_space_proj_prior_mean.numpy(),
                atol=1e-5,
                rtol=1e-5,
            )

            npt.assert_allclose(
                null_space_proj_cov.numpy(),
                null_space_proj_prior_cov.numpy(),
                atol=1e-5,
                rtol=1e-5,
            )


@pytest.mark.parametrize(
    "cov",
    [
        params.FactorizedCovariance(),
        params.DiagonalCovariance(),
        params.LowRankCovariance(rank=2),
        params.KroneckerCovariance(input_rank=2, output_rank=3),
    ],
    ids=lambda x: x.__class__.__name__,
)
def test_if_bias_is_none_cov_bias_is_none(cov):
    """Test whether if the bias is None the covariance doesn't have bias parameters."""
    linear_layer = bnn.Linear(5, 3, bias=False, cov=cov)

    for name, param in linear_layer.params.cov.named_parameters():
        if "bias" in name:
            assert param is None


@pytest.mark.parametrize(
    "parametrization",
    [
        params.Standard(),
        params.NeuralTangent(),
        params.MaximalUpdate(),
    ],
    ids=lambda x: x.__class__.__name__,
)
@pytest.mark.parametrize("layer_type", ["input", "hidden", "output"], ids=lambda x: x)
def test_weight_initialization(parametrization, layer_type):
    """Test whether the initialization of the weights is correct."""
    torch.manual_seed(2452345)

    with torch.no_grad():
        linear_layer = bnn.Linear(
            333,
            333,
            layer_type=layer_type,
            parametrization=parametrization,
            cov=None,
        )

        fan_in, fan_out = nn.init._calculate_fan_in_and_fan_out(
            linear_layer.params.weight
        )

        npt.assert_allclose(
            linear_layer.weight.mean(),
            0.0,
            atol=1e-2,
            rtol=1e-2,
        )

        npt.assert_allclose(
            linear_layer.weight.std(),
            parametrization.weight_init_scale(
                fan_in=fan_in, fan_out=fan_out, layer_type=layer_type
            ),
            atol=1e-2,
            rtol=1e-2,
        )


@pytest.mark.parametrize(
    "parametrization",
    [
        params.Standard(),
        params.NeuralTangent(),
        params.MaximalUpdate(),
    ],
    ids=lambda x: x.__class__.__name__,
)
@pytest.mark.parametrize("layer_type", ["input", "hidden", "output"], ids=lambda x: x)
def test_bias_initialization(parametrization, layer_type):
    """Test whether the initialization of the bias is correct."""
    torch.manual_seed(2452345)

    with torch.no_grad():
        linear_layer = bnn.Linear(
            333,
            333,
            bias=True,
            layer_type=layer_type,
            parametrization=parametrization,
            cov=None,
        )

        fan_in, fan_out = nn.init._calculate_fan_in_and_fan_out(
            linear_layer.params.weight
        )

        npt.assert_allclose(
            linear_layer.bias.mean(),
            0.0,
            atol=1e-1,
            rtol=1e-1,
        )

        npt.assert_allclose(
            linear_layer.bias.std(),
            parametrization.bias_init_scale(
                fan_in=fan_in, fan_out=fan_out, layer_type=layer_type
            ),
            atol=1e-2,
            rtol=1e-2,
        )


@pytest.mark.parametrize(
    "parametrization",
    [
        params.Standard(),
        params.NeuralTangent(),
        params.MaximalUpdate(),
    ],
    ids=lambda x: x.__class__.__name__,
)
@pytest.mark.parametrize("layer_type", ["input", "hidden", "output"], ids=lambda x: x)
@pytest.mark.parametrize(
    "layer_width,cov",
    [
        (33, params.FactorizedCovariance()),
        (100, params.LowRankCovariance(5)),
        (100, params.KroneckerCovariance()),
        (100, params.DiagonalCovariance()),
    ],
    ids=lambda x: (
        x.__class__.__name__
        if isinstance(x, params.FactorizedCovariance)
        else f"width_{x}"
    ),
)
def test_covariance_parameter_initialization(
    parametrization, layer_type, layer_width, cov
):
    """Test whether the initialization of the covariance parameters is correct."""
    torch.manual_seed(2452345)

    with torch.no_grad():
        linear_layer = bnn.Linear(
            layer_width,
            layer_width,
            bias=True,
            layer_type=layer_type,
            parametrization=parametrization,
            cov=cov,
        )

        fan_in, fan_out = nn.init._calculate_fan_in_and_fan_out(
            linear_layer.params.weight
        )

        if isinstance(cov, params.DiagonalCovariance):
            covariance_specific_scaling_factor = 1.0
        elif isinstance(cov, params.KroneckerCovariance):
            covariance_specific_scaling_factor = 1 / math.sqrt(cov.input_rank)
        elif isinstance(cov, params.LowRankCovariance):
            covariance_specific_scaling_factor = 1 / math.sqrt(cov.rank)
        elif isinstance(cov, params.FactorizedCovariance):
            covariance_specific_scaling_factor = 1 / math.sqrt(cov.rank)
        else:
            raise NotImplementedError()

        for name, param in linear_layer.params.cov.named_parameters():
            if param is not None:
                npt.assert_allclose(
                    param.mean(),
                    0.0,
                    atol=1e-1,
                    rtol=1e-1,
                )
            if "weight" in name:
                npt.assert_allclose(
                    param.std(),
                    parametrization.weight_init_scale(
                        fan_in=fan_in, fan_out=fan_out, layer_type=layer_type
                    )
                    * covariance_specific_scaling_factor,
                    atol=1e-1,
                    rtol=1e-1,
                )
            elif "bias" in name:
                npt.assert_allclose(
                    param.std(),
                    parametrization.bias_init_scale(
                        fan_in=fan_in, fan_out=fan_out, layer_type=layer_type
                    )
                    * covariance_specific_scaling_factor,
                    atol=1e-1,
                    rtol=1e-1,
                )



from .covariances.diagonal import DiagonalCovariance
from .covariances.factorized import FactorizedCovariance
from .covariances.kronecker import KroneckerCovariance
from .covariances.low_rank import LowRankCovariance
from .gaussian_parameter import GaussianParameter
from .parameter import BNNParameter
from .parametrizations import MaximalUpdate, NeuralTangent, Parametrization, Standard

SP = Standard
NTP = NeuralTangent
MUP = MaximalUpdate



from __future__ import annotations

import math
from typing import TYPE_CHECKING

import numpy as np
import torch
from torch import nn

if TYPE_CHECKING:
    from jaxtyping import Float
    from torch import Tensor


class FactorizedCovariance(nn.Module):
    r"""Covariance of a Gaussian parameter with a factorized structure.

    Assumes the covariance is factorized as a product of a square matrix and its transpose.

    ..math::
        \mathbf{\Sigma} = \mathbf{S} \mathbf{S}^\top

    :param rank: Rank of the covariance matrix. If `None`, the rank is set to the total number
        of mean parameters.
    """

    def __init__(self, rank: int | None = None):
        super().__init__()
        if rank is not None and rank < 1:
            raise ValueError(f"Rank must be at least 1, but is {rank} < 1.")
        self.rank = rank

    def initialize_parameters(
        self,
        mean_parameters: dict[str, torch.Tensor],
    ) -> None:
        """Initialize the covariance parameters.

        :param mean_parameters: Mean parameters of the Gaussian distribution.
        :return: Covariance parameters.
        """
        # Total number of mean parameters
        numel_mean_parameters = sum(
            [tens.numel() for tens in mean_parameters.values() if tens is not None]
        )
        if self.rank is None or self.rank > numel_mean_parameters:
            self.rank = numel_mean_parameters

        # Initialize factor parameters
        self.factor = nn.ParameterDict(
            {
                name: (
                    torch.empty(
                        *param.shape,
                        self.rank,
                        dtype=param.dtype,
                        device=param.device,
                    )
                    if param is not None
                    else None
                )
                for name, param in mean_parameters.items()
            }
        )

    def reset_parameters(
        self,
        mean_parameter_scales: dict[str, float] | float = 1.0,
    ) -> None:
        """Reset the parameters of the covariance matrix.

        Initalizes the parameters of the covariance matrix with a
        scale that is given by the mean parameter scales and a
        covariance-specific scaling that depends on the structure of the covariance matrix.

        :param mean_parameter_scales: Scales of the mean parameters. If a dictionary
            keys are the names of the mean parameters. If a float, all covariance
            parameters are initialized with the same scale.
        """
        if isinstance(mean_parameter_scales, float):
            mean_parameter_scales = {
                name: mean_parameter_scales for name in self.factor.keys()
            }

        for name, param in self.factor.items():
            if param is not None:
                nn.init.normal_(
                    param,
                    mean=0,
                    std=mean_parameter_scales[name] / math.sqrt(self.rank),
                )

    def factor_matmul(
        self,
        input: Float[Tensor, "*sample parameter"],
        /,
        additive_constant: Float[Tensor, "*sample parameter"] | None = None,
    ) -> dict[str, Float[Tensor, "*sample parameter"]]:
        """Multiply left factor of the covariance matrix with the input.

        :param input: Input tensor.
        :param additive_constant: Additive constant to be added to the output.
        """
        sample_shape = input.shape[:-1]

        # Stack factor parameters
        stacked_parameters = self._stacked_parameters()

        # Multiply factor parameters with input
        result = torch.einsum("...p,qp->...q", input, stacked_parameters)

        if additive_constant is not None:
            result = result + additive_constant

        # Split result into parameter shapes
        split_result = torch.tensor_split(
            result,
            list(
                np.cumsum(
                    [
                        tens[..., 0].numel()
                        for tens in self.factor.values()
                        if tens is not None
                    ]
                )[:-1]
            ),
            dim=-1,
        )

        result_dict = {}
        i = 0
        for name, param in self.factor.items():
            if param is not None:
                result_dict[name] = split_result[i].view(
                    *sample_shape, *param.shape[:-1]
                )
                i += 1

        return result_dict

    def _stacked_parameters(self) -> Float[Tensor, "parameter"]:
        """Stack parameters into a single tensor."""
        stacked_parameters = torch.vstack(
            [
                factor_param.view(-1, self.rank)
                for factor_param in self.factor.values()
                if factor_param is not None
            ]
        )
        return stacked_parameters

    def to_dense(self) -> Float[Tensor, "parameter parameter"]:
        """Convert the covariance matrix to a dense representation."""
        stacked_parameters = self._stacked_parameters()
        return stacked_parameters @ stacked_parameters.mT

    @property
    def lr_scaling(self) -> dict[str, float]:
        """Compute the learning rate scaling for the covariance parameters."""
        return {"factor." + name: 1 / self.rank for name in self.factor.keys()}



from __future__ import annotations

from typing import TYPE_CHECKING, Literal

import torch
from torch import nn

from .. import params
from .module import BNNModule

if TYPE_CHECKING:
    from jaxtyping import Float
    from torch import Tensor


class Linear(BNNModule):
    """
    Applies an affine transformation to the input.

    :param in_features: Size of each input sample.
    :param out_features: Size of each output sample.
    :param bias: If set to ``False``, the layer will not learn an additive bias.
    :param layer_type: Type of the layer. Can be one of "input", "hidden", or "output".
        Controls the initialization and learning rate scaling of the parameters.
    :param cov: Covariance object for the parameters.
    :param parametrization: The parametrization to use. Defines the initialization
        and learning rate scaling for the parameters of the module.
    :param device: Device on which to create the parameters.
    :param dtype: Data type of the parameters.
    """

    def __init__(
        self,
        in_features: int,
        out_features: int,
        bias: bool = True,
        layer_type: Literal["input", "hidden", "output"] = "hidden",
        cov: params.FactorizedCovariance | None = None,
        parametrization: params.Parametrization = params.MaximalUpdate(),
        device: torch.device | None = None,
        dtype: torch.dtype | None = None,
    ):
        factory_kwargs = {"device": device, "dtype": dtype}
        super().__init__(parametrization=parametrization)

        self.in_features = in_features
        self.out_features = out_features

        if layer_type not in ["input", "hidden", "output"]:
            raise ValueError(
                f"Invalid layer type '{layer_type}'. Must be one of ['input', 'hidden', 'output']."
            )
        self.layer_type = layer_type

        # Weight and bias parameters
        mean_param_dict = {
            **{"weight": torch.empty((out_features, in_features), **factory_kwargs)},
            **(
                {"bias": torch.empty(out_features, **factory_kwargs)}
                if bias
                else {"bias": None}
            ),
        }
        if cov is None:
            self.params = nn.ParameterDict(mean_param_dict)
            self.params.cov = None
        else:
            self.params = params.GaussianParameter(
                mean=mean_param_dict,
                cov=cov,
            )

        # Temperature parameter
        if layer_type == "output":
            self.params.temperature = nn.Parameter(
                torch.ones(1, **factory_kwargs), requires_grad=False
            )

        self.reset_parameters()

    @property
    def weight(self) -> nn.Parameter:
        return self.params.weight

    @property
    def bias(self) -> nn.Parameter:
        return self.params.bias

    def reset_parameters(
        self,
    ) -> None:
        """Reset the parameters of the module."""
        fan_in, fan_out = nn.init._calculate_fan_in_and_fan_out(self.params.weight)

        # Mean parameters
        mean_parameter_scales = {}
        mean_parameter_scales["weight"] = self.parametrization.weight_init_scale(
            fan_in, fan_out, layer_type=self.layer_type
        )
        nn.init.normal_(self.params.weight, mean=0, std=mean_parameter_scales["weight"])

        if self.params.bias is not None:
            mean_parameter_scales["bias"] = self.parametrization.bias_init_scale(
                fan_in, fan_out, layer_type=self.layer_type
            )
            nn.init.normal_(self.params.bias, mean=0, std=mean_parameter_scales["bias"])

        # Covariance parameters
        if self.params.cov is not None:
            self.params.cov.reset_parameters(mean_parameter_scales)

        # Temperature parameter
        if hasattr(self.params, "temperature"):
            nn.init.constant_(self.params.temperature, 1.0)
            self.params.temperature.requires_grad = False

    def parameters_and_lrs(
        self,
        lr: float,
        optimizer: Literal["SGD", "Adam"] = "SGD",
    ) -> list[dict[str, Tensor | float]]:
        fan_in, fan_out = nn.init._calculate_fan_in_and_fan_out(self.params.weight)

        mean_parameter_lr_scales = {}
        mean_parameter_lr_scales["weight"] = self.parametrization.weight_lr_scale(
            fan_in, fan_out, optimizer=optimizer, layer_type=self.layer_type
        )
        param_groups = [
            {
                "params": self.params.weight,
                "lr": lr * mean_parameter_lr_scales["weight"],
            }
        ]

        if self.params.bias is not None:
            mean_parameter_lr_scales["bias"] = self.parametrization.bias_lr_scale(
                fan_in, fan_out, optimizer=optimizer, layer_type=self.layer_type
            )
            param_groups += [
                {
                    "params": self.params.bias,
                    "lr": lr * mean_parameter_lr_scales["bias"],
                }
            ]

        if self.params.cov is not None:
            for name, param in self.params.cov.named_parameters():
                lr_scaling = 1.0
                if "weight" in name:
                    lr_scaling = mean_parameter_lr_scales["weight"]
                elif "bias" in name:
                    lr_scaling = mean_parameter_lr_scales["bias"]
                param_groups += [
                    {
                        "params": param,
                        "lr": lr * lr_scaling * self.params.cov.lr_scaling[name],
                    }
                ]

        return param_groups

    def forward(
        self,
        input: Float[Tensor, "*sample *batch in_feature"],
        /,
        sample_shape: torch.Size = torch.Size([]),
        generator: torch.Generator | None = None,
        input_contains_samples: bool = False,
        parameter_samples: dict[str, Float[Tensor, "*sample parameter"]] | None = None,
    ) -> Float[Tensor, "*sample *batch out_feature"]:

        if (parameter_samples is None) and (self.params.cov is None):
            output = nn.functional.linear(input, self.params.weight, self.params.bias)

            # Scale with inverse temperature if not training and the parameters are in the output layer
            if hasattr(self.params, "temperature") and not self.training:
                output = output / self.params.temperature

            if not input_contains_samples:
                # Repeat output for each desired sample
                output = output.expand(*sample_shape, *output.shape)

        else:
            if not input_contains_samples:
                # Repeat input for each desired sample
                input = input.expand(*sample_shape, *input.shape)

            batch_shape = input.shape[len(sample_shape) : -1]

            if parameter_samples is None:
                # Sample one set of parameters per sample
                parameter_samples = self.params.sample(
                    sample_shape=sample_shape, generator=generator
                )

            # Affine transformation of input via sampled weights and biases
            weight_samples = parameter_samples["weight"]
            sample_dim_idcs = list(range(len(sample_shape)))
            output = torch.einsum(
                weight_samples,
                [*sample_dim_idcs, len(sample_dim_idcs) + 1, len(sample_dim_idcs) + 2],
                input,
                [*sample_dim_idcs, ..., len(sample_dim_idcs) + 2],
                [
                    *sample_dim_idcs,
                    ...,
                    len(sample_dim_idcs) + 1,
                ],  # Ensure sample dimensions are in front, batch dimensions next and out_feature dimensions last
            )
            if "bias" in parameter_samples:
                bias_samples = parameter_samples["bias"]
                output = output + bias_samples.view(
                    *(
                        sample_shape
                        + len(batch_shape) * (1,)
                        + (bias_samples.shape[-1],)
                    )
                )

        return output

    def extra_repr(self) -> str:
        return f"in_features={self.in_features}, out_features={self.out_features}, bias={self.params.bias is not None}"

    def _load_from_state_dict(
        self,
        state_dict: dict,
        prefix: str,
        local_metadata: dict,
        strict: bool,
        missing_keys: list[str],
        unexpected_keys: list[str],
        error_msgs: list[str],
    ):

        # Ensure compatibility with nn.Linear
        if prefix + "weight" in state_dict:
            state_dict[prefix + "params.weight"] = state_dict.pop(prefix + "weight")
        if prefix + "bias" in state_dict:
            state_dict[prefix + "params.bias"] = state_dict.pop(prefix + "bias")
        if (
            hasattr(self.params, "temperature")
            and prefix + "params.temperature" not in state_dict
        ):
            state_dict[prefix + "params.temperature"] = self.params.temperature.data

        super()._load_from_state_dict(
            state_dict,
            prefix,
            local_metadata,
            strict,
            missing_keys,
            unexpected_keys,
            error_msgs,
        )



import copy

import numpy as np
from numpy import testing as npt
import torch

from inferno.bnn import params

import pytest


@pytest.mark.parametrize(
    "mean",
    [
        {"weight": torch.randn(4, 2), "bias": torch.randn(4)},
        {"weight": torch.randn(4, 2), "bias": None},
        {"weight": torch.randn(1, 1)},
        {"weight": torch.randn(3, 2, 1, 2), "bias": torch.randn((3,))},
    ],
)
@pytest.mark.parametrize(
    "cov",
    [
        params.FactorizedCovariance(),
        params.DiagonalCovariance(),
        params.LowRankCovariance(rank=2),
        params.KroneckerCovariance(input_rank=2, output_rank=3),
    ],
    ids=lambda x: x.__class__.__name__,
)
def test_sample(mean, cov):
    """Test sampling from a Gaussian parameter."""
    torch.random.manual_seed(463)

    # Initialize the Gaussian parameter
    cov = copy.deepcopy(cov)
    param = params.GaussianParameter(mean, cov)
    param.cov.reset_parameters()

    # Sample
    sample_shape = torch.Size((100000,))
    parameter_samples = param.sample(sample_shape)

    for name, tens in parameter_samples.items():
        if param[name] is None:
            assert tens is None
        else:
            # Check that the shape is correct
            assert tens.shape == sample_shape + param[name].shape

            # Check that the mean is correct
            npt.assert_allclose(
                tens.mean(dim=tuple(np.arange(len(sample_shape)))).detach().numpy(),
                param[name].detach().numpy(),
                rtol=1e-1,
                atol=1e-1,
            )

    # Check that the covariance is correct
    npt.assert_allclose(
        torch.hstack(
            [tens.view(*sample_shape, -1) for tens in parameter_samples.values()]
        )
        .mT.cov()
        .detach()
        .numpy(),
        param.cov.to_dense().detach().numpy(),
        rtol=1e-1,
        atol=1e-1,
    )



from __future__ import annotations

import math
from typing import TYPE_CHECKING

import torch
from torch import nn

from .factorized import FactorizedCovariance

if TYPE_CHECKING:
    from jaxtyping import Float
    from torch import Tensor


class KroneckerCovariance(FactorizedCovariance):
    r"""Covariance of a Gaussian parameter with Kronecker structure.

    Assumes the covariance is given by a Kronecker product of two matrices of size
    equal to the number of inputs and outputs to the layer. Each Kronecker factor is
    assumed to be of rank $R \leq D$ where $D$ is either the input or
    output dimension of the layer.

    More precisely, the covariance is given by

    $$
    \begin{align*}
        \mathbf{\Sigma} &= \mathbf{C}_{\text{in}} \otimes \mathbf{C}_{\text{out}}\\
                        &= \mathbf{S}_{\text{in}}\mathbf{S}_{\text{in}}^\top \otimes \mathbf{S}_{\text{out}}\mathbf{S}_{\text{out}}^\top\\
                        &= (\mathbf{S}_{\text{in}} \otimes \mathbf{S}_{\text{out}}) (\mathbf{S}_{\text{in}}^\top \otimes \mathbf{S}_{\text{out}}^\top)
    \end{align*}
    $$

    where $\mathbf{S}_{\text{in}}$ and $\mathbf{S}_{\text{out}}$ are the
    low-rank factors of the Kronecker factors $\mathbf{C}_{\text{in}}$ and 
    $\mathbf{C}_{\text{out}}$.

    :param input_rank: Rank of the input Kronecker factor. If None, assumes full rank.
    :param output_rank: Rank of the output Kronecker factor. If None, assumes full rank.
    """

    def __init__(
        self,
        input_rank: int | None = None,
        output_rank: int | None = None,
    ):

        if input_rank is not None and input_rank < 1:
            raise ValueError(f"Input rank must be at least 1, but is {input_rank} < 1.")
        if output_rank is not None and output_rank < 1:
            raise ValueError(
                f"Output rank must be at least 1, but is {output_rank} < 1."
            )
        self.input_rank = input_rank
        self.output_rank = output_rank

        # rank(A \otimes B) = rank(A) * rank(B) where A and B are the Kronecker factors
        super().__init__(
            rank=(
                None
                if (input_rank is None) or (output_rank is None)
                else input_rank * output_rank
            )
        )

    def initialize_parameters(
        self,
        mean_parameters: dict[str, torch.Tensor],
    ) -> None:

        # Sizes of the Kronecker factors
        out_dim = mean_parameters["weight"].shape[0]
        in_dim = math.prod(mean_parameters["weight"].shape[1:])
        if "bias" in mean_parameters.keys() and mean_parameters["bias"] is not None:
            in_dim += mean_parameters["bias"].ndim
        self._fan_in = nn.init._calculate_fan_in_and_fan_out(mean_parameters["weight"])[
            0
        ]

        # Set the Kronecker factor ranks if not set
        if self.input_rank is None:
            self.input_rank = in_dim
        if self.output_rank is None:
            self.output_rank = out_dim
        if self.input_rank > in_dim:
            self.input_rank = in_dim
        if self.output_rank > out_dim:
            self.output_rank = out_dim

        # rank(A \otimes B) = rank(A) * rank(B) where A and B are the Kronecker factors
        self.rank = self.input_rank * self.output_rank

        # Initialize factor parameters
        input_factor_dict = {}
        for name, param in mean_parameters.items():
            if name == "bias":
                if mean_parameters["bias"] is not None:
                    input_factor_dict[name] = torch.empty(
                        1, self.input_rank, dtype=param.dtype, device=param.device
                    )
                else:
                    input_factor_dict[name] = None
            elif name == "weight":
                # Shapes (especially for convolutions) are inspired by K-FAC
                # See https://fdangel.com/posts/kfac_explained.html
                input_factor_dict[name] = torch.empty(
                    *param.shape[1:],
                    self.input_rank,
                    dtype=param.dtype,
                    device=param.device,
                )
                self.output_factor = nn.Parameter(
                    torch.empty(
                        param.shape[0],
                        self.output_rank,
                        dtype=param.dtype,
                        device=param.device,
                    )
                )
            else:
                raise NotImplementedError(
                    "Kronecker covariance currently only supports bias and weight parameters."
                )

        self.input_factor = nn.ParameterDict(input_factor_dict)

    def reset_parameters(
        self,
        mean_parameter_scales: dict[str, float] | float = 1.0,
    ) -> None:
        if isinstance(mean_parameter_scales, float):
            mean_parameter_scales = {
                name: mean_parameter_scales for name in self.input_factor.keys()
            }

        for name, param in self.input_factor.items():
            if param is not None:
                nn.init.normal_(
                    param,
                    mean=0.0,
                    std=mean_parameter_scales[name] / math.sqrt(self.input_rank),
                    # * math.pow(self._fan_in, 1 / 4),
                    # Last factor cancels with scaling in output_factor in forward pass,
                    # but approximately distributes mean parameter scaling across in- and output factor
                )

        nn.init.normal_(
            self.output_factor,
            mean=0.0,
            std=1 / math.sqrt(self.output_rank),  # * math.pow(self._fan_in, -1 / 4),
            # Last factor cancels with scaling in input_factor in forward pass,
            # but approximately distributes mean parameter scaling across in- and output factor),
        )

    def factor_matmul(
        self,
        input: Float[Tensor, "*sample parameter"],
        /,
        additive_constant: Float[Tensor, "*sample parameter"] | None = None,
    ) -> dict[str, Float[Tensor, "*sample parameter"]]:
        sample_shape = input.shape[:-1]

        # Stack factor parameters (input x rank, output x rank)
        stacked_input_factor, stacked_output_factor = self._stacked_parameters()

        # Multiply Kronecker factors with input
        # (A \otimes B) colvec(X) = colvec(B X A')
        result = stacked_output_factor @ (
            input.view(*sample_shape, self.input_rank, self.output_rank).mT
            @ stacked_input_factor.mT
        )
        result = result.mT.reshape(*sample_shape, -1)

        if additive_constant is not None:
            result = result + additive_constant

        # Split result into parameter shapes
        result_dict = {}
        starting_idx_weight_result = 0
        if "bias" in self.input_factor.keys() and self.input_factor["bias"] is not None:
            result_dict["bias"] = result[..., : self.output_factor.shape[0]]
            starting_idx_weight_result = self.output_factor.shape[0]
        if "weight" in self.input_factor.keys():
            result_dict["weight"] = result[..., starting_idx_weight_result:].view(
                *sample_shape,
                self.output_factor.shape[0],
                *self.input_factor["weight"].shape[:-1],
            )
        return result_dict

    @property
    def sample_scale(self) -> float:
        return 1.0 / math.sqrt(self.rank)

    def _stacked_parameters(self) -> tuple[Float[Tensor, "parameter"]]:
        stacked_input_factor = torch.vstack(
            [
                factor_param.view(-1, self.input_rank)
                for factor_param in self.input_factor.values()
                if factor_param is not None
            ]
        )
        return stacked_input_factor, self.output_factor

    def to_dense(self) -> Float[Tensor, "parameter parameter"]:
        stacked_parameters_input_factor, stacked_parameters_output_factor = (
            self._stacked_parameters()
        )
        return torch.kron(
            stacked_parameters_input_factor @ stacked_parameters_input_factor.mT,
            stacked_parameters_output_factor @ stacked_parameters_output_factor.mT,
        )

    @property
    def lr_scaling(self) -> dict[str, float]:
        """Compute the learning rate scaling for the covariance parameters."""
        return {
            **{
                "input_factor." + name: 1 / self.input_rank
                for name in self.input_factor.keys()
            },
            **{
                "output_factor": 1 / self.output_rank,
            },
        }



from __future__ import annotations

import abc
from typing import TYPE_CHECKING

import torch
from torch import nn

if TYPE_CHECKING:
    from jaxtyping import Float
    from torch import Tensor


class BNNParameter(nn.ParameterDict, abc.ABC):

    def __init__(
        self, hyperparameters: dict[str, Float[Tensor, "*hyperparameter"]] | None = None
    ):
        super().__init__(hyperparameters)

    def sample(
        self,
        sample_shape: torch.Size = torch.Size([]),
        generator: torch.Generator | None = None,
    ) -> Float[Tensor, "*sample parameter"]:
        raise NotImplementedError



import numpy.testing as npt
import torch
from torch import nn

import inferno
from inferno import bnn, loss_fns, models
from inferno.bnn import params

import pytest


@pytest.mark.parametrize(
    "model,loss_fn,inputs,targets",
    [
        (
            bnn.Sequential(
                models.MLP(
                    in_size=100,
                    hidden_sizes=[64, 64],
                    out_size=1,
                    cov=params.DiagonalCovariance(),
                ),
                nn.Flatten(-2, -1),
            ),
            nn.BCEWithLogitsLoss,
            torch.randn(32, 100),
            torch.randint(0, 10, (32,)).float(),
        ),
        (
            models.MLP(
                in_size=100,
                hidden_sizes=[64, 64],
                out_size=10,
                cov=params.DiagonalCovariance(),
            ),
            nn.CrossEntropyLoss,
            torch.randn(32, 100),
            torch.randint(0, 1, (32,)),
        ),
        (
            models.ResNet18(
                out_size=10,
                architecture="cifar",
                cov=params.DiagonalCovariance(),
            ),
            nn.CrossEntropyLoss,
            torch.randn(32, 3, 32, 32),
            torch.randint(0, 10, (32,)),
        ),
    ],
    ids=lambda x: x.__class__.__name__,
)
def test_vfe_for_different_models(model, loss_fn, inputs, targets):
    """Test whether the VFE loss function works for different models."""
    torch.manual_seed(0)

    for variational_loss_fn in [loss_fns.VariationalFreeEnergy, loss_fns.NegativeELBO]:
        variational_free_energy = variational_loss_fn(
            nll=loss_fn(),
            model=model,
        )

        variational_free_energy(model(inputs), targets)



import math

from numpy import testing as npt
import torch
from torch import nn

from inferno import bnn
from inferno.bnn import params

import pytest


@pytest.mark.parametrize(
    "parametrization",
    [params.Standard, params.NeuralTangent, params.MaximalUpdate],
)
def test_changing_parametrization_of_bnn_module_changes_parametrization_of_children(
    parametrization,
):
    """When defining a BNNModule with a certain parametrization, changing the parametrization
    of the module should change the parametrization of all children that are BNNModules.
    """

    class MyModel(bnn.BNNModule):

        def __init__(self):
            super().__init__(parametrization=params.Standard())
            self.linear1 = bnn.Linear(
                2,
                5,
                layer_type="input",
                bias=True,
            )
            self.linear2 = bnn.Linear(
                5,
                3,
                bias=False,
                cov=params.FactorizedCovariance(),
            )
            self.linear3 = nn.Linear(3, 1, bias=True)
            self.relu1 = nn.ReLU()
            self.relu2 = nn.ReLU()
            self.relu3 = nn.ReLU()

            self.reset_parameters()

        def forward(self, input):
            input = self.linear1(input)
            input = self.relu1(input)
            input = self.linear2(input)
            input = self.relu2(input)
            input = self.linear3(input)
            input = self.relu3(input)
            return input

    model = MyModel()

    # Check initial parametrization
    for child in model.children():
        if isinstance(child, bnn.BNNModule):
            assert isinstance(child.parametrization, params.Standard)

    # Change parametrization and check
    model.parametrization = parametrization()

    for child in model.children():
        if isinstance(child, bnn.BNNModule):
            assert isinstance(child.parametrization, parametrization)



from __future__ import annotations

import abc
from typing import Literal

LayerType = Literal["input", "hidden", "output"]
ParamType = Literal["weight", "bias"]


class Parametrization(abc.ABC):
    """Abstract base class for all neural network parametrizations."""

    def weight_init_scale(
        self,
        fan_in: int,
        fan_out: int,
        layer_type: LayerType = "hidden",
    ) -> float:
        """Compute the weight initialization scale for the given layer.

        :param fan_in: Number of inputs to the layer.
        :param fan_out: Number of outputs from the layer.
        :param layer_type: Type of the layer. Can be one of "input", "hidden", or "output".
        """
        raise NotImplementedError

    def bias_init_scale(
        self,
        fan_in: int,
        fan_out: int,
        layer_type: LayerType = "hidden",
    ) -> float:
        """Compute the bias initialization scale for the given layer.

        :param fan_in: Number of inputs to the layer.
        :param fan_out: Number of outputs from the layer.
        :param layer_type: Type of the layer. Can be one of "input", "hidden", or "output".
        """
        raise NotImplementedError

    def weight_lr_scale(
        self,
        fan_in: int,
        fan_out: int,
        optimizer: str,
        layer_type: LayerType = "hidden",
    ) -> float:
        """Compute the learning rate scale for the weight parameters.

        :param fan_in: Number of inputs to the layer.
        :param fan_out: Number of outputs from the layer.
        :param optimizer: Optimizer being used.
        :param layer_type: Type of the layer. Can be one of "input", "hidden", or "output".
        """
        raise NotImplementedError

    def bias_lr_scale(
        self,
        fan_in: int,
        fan_out: int,
        optimizer: str,
        layer_type: LayerType = "hidden",
    ) -> float:
        """Compute the learning rate scale for the bias parameters.

        :param fan_in: Number of inputs to the layer.
        :param fan_out: Number of outputs from the layer.
        :param optimizer: Optimizer being used.
        :param layer_type: Type of the layer. Can be one of "input", "hidden", or "output".
        """
        raise NotImplementedError



import math

import numpy.testing as npt
import torch
from torch import nn

from inferno import bnn
from inferno.bnn import params

import pytest


@pytest.mark.parametrize("batch_shape", [(), (1,), (4,)])
@pytest.mark.parametrize("in_channels, out_channels", [(1, 1), (1, 2)])
@pytest.mark.parametrize(
    "in_features, kernel_size", [((1, 1, 2), 1), ((4, 2, 6), (2, 1, 2))]
)
@pytest.mark.parametrize("cov", [None, bnn.params.FactorizedCovariance()])
def test_generalizes_pytorch_conv3d_layer(
    batch_shape,
    in_channels,
    out_channels,
    in_features,
    kernel_size,
    cov,
):
    """Test whether the BNN Conv3d layer generalizes the PyTorch Conv3d layer."""
    generator = torch.Generator().manual_seed(0)

    input = torch.randn(
        batch_shape + (in_channels, *in_features),
        generator=generator,
    )
    conv_layer = torch.nn.Conv3d(
        in_channels,
        out_channels,
        kernel_size=kernel_size,
    )
    conv_layer.weight = torch.nn.Parameter(conv_layer.weight)
    conv_layer.bias = torch.nn.Parameter(conv_layer.bias)

    bnn_conv_layer = bnn.Conv3d(
        in_channels,
        out_channels,
        kernel_size=kernel_size,
        cov=cov,
    )
    bnn_conv_layer.params.weight = torch.nn.Parameter(conv_layer.weight)
    bnn_conv_layer.params.bias = torch.nn.Parameter(conv_layer.bias)
    if bnn_conv_layer.params.cov is not None:
        for name, param in bnn_conv_layer.params.cov.named_parameters():
            torch.nn.init.constant_(param, 0.0)

    npt.assert_allclose(
        conv_layer(input).detach().numpy(),
        bnn_conv_layer(input, generator=generator).detach().numpy(),
        atol=1e-6,
        rtol=1e-6,
    )


@pytest.mark.parametrize("sample_shape", [(), (1,), (2,), (3, 2)])
@pytest.mark.parametrize("batch_shape", [(), (1,), (3,)])
@pytest.mark.parametrize("in_features", [(2, 4, 4), (1, 5, 1)])
@pytest.mark.parametrize(
    "conv3d",
    [
        bnn.Conv3d(in_channels=3, out_channels=1, kernel_size=(1, 2, 1)),
        bnn.Conv3d(in_channels=1, out_channels=2, bias=False, kernel_size=1),
    ],
)
def test_shape(sample_shape, batch_shape, in_features, conv3d):
    """Test whether the output shape is correct."""
    generator = torch.Generator().manual_seed(0)
    input = torch.randn(
        *batch_shape, conv3d.in_channels, *in_features, generator=generator
    )
    output = conv3d(input, sample_shape=sample_shape, generator=generator)

    assert output.shape == (
        *sample_shape,
        *nn.Conv3d(
            in_channels=conv3d.in_channels,
            out_channels=conv3d.out_channels,
            kernel_size=conv3d.kernel_size,
        )(input).shape,
    )


@pytest.mark.parametrize("seed", [2576, 897, 999])
def test_forward_is_deterministic_given_generator(seed):
    """Test whether the forward method is deterministic given a generator."""
    conv_layer = bnn.Conv3d(5, 3, kernel_size=2)

    input = torch.randn(
        5, 9, 2, 4, generator=torch.Generator().manual_seed(seed + 2452345)
    )
    output1 = conv_layer(input, generator=torch.Generator().manual_seed(seed))
    output2 = conv_layer(input, generator=torch.Generator().manual_seed(seed))

    npt.assert_allclose(output1.detach().numpy(), output2.detach().numpy())


@pytest.mark.parametrize(
    "conv_layer_to_load",
    [
        bnn.Conv3d(3, 2, kernel_size=(1, 2, 2)),
        bnn.Conv3d(3, 2, kernel_size=(1, 2, 2), cov=params.FactorizedCovariance()),
        nn.Conv3d(3, 2, kernel_size=2),
    ],
)
def test_load_from_state_dict(conv_layer_to_load):
    """Test whether the load_from_state_dict method is working for torch and inferno
    Conv3d layers."""
    state_dict = conv_layer_to_load.state_dict()
    new_conv_layer = bnn.Conv3d(
        conv_layer_to_load.in_channels,
        conv_layer_to_load.out_channels,
        kernel_size=conv_layer_to_load.kernel_size,
        cov=params.FactorizedCovariance(),
    )
    new_conv_layer.load_state_dict(
        state_dict,
        strict=hasattr(conv_layer_to_load, "params.cov"),
    )

    prefix = "params." if isinstance(conv_layer_to_load, bnn.BNNModule) else ""

    npt.assert_allclose(
        new_conv_layer.params.weight.detach().numpy(),
        state_dict[prefix + "weight"].detach().numpy(),
    )
    npt.assert_allclose(
        new_conv_layer.params.bias.detach().numpy(),
        state_dict[prefix + "bias"].detach().numpy(),
    )

    if hasattr(conv_layer_to_load, "params.cov"):
        npt.assert_allclose(
            new_conv_layer.params.cov_params.detach().numpy(),
            state_dict[prefix + "cov_params"].detach().numpy(),
        )


def test_register_forward_hook():
    """Test whether the register_forward_hook method is working."""
    generator = torch.Generator().manual_seed(45)

    conv_layer = bnn.Conv3d(3, 2, kernel_size=1)
    test_dict = {"hook_has_fired": False}

    def hook(module, input, output):
        test_dict["hook_has_fired"] = True

    conv_layer.register_forward_hook(hook)
    conv_layer(torch.randn(5, 3, 9, 4, 4, generator=generator))

    assert test_dict["hook_has_fired"]


@pytest.mark.parametrize(
    "cov",
    [
        params.FactorizedCovariance(),
        params.DiagonalCovariance(),
        params.LowRankCovariance(rank=2),
        params.KroneckerCovariance(input_rank=2, output_rank=3),
    ],
    ids=lambda x: x.__class__.__name__,
)
def test_if_bias_is_none_cov_bias_is_none(cov):
    """Test whether if the bias is None the covariance doesn't have bias parameters."""
    linear_layer = bnn.Conv3d(5, 3, kernel_size=5, bias=False, cov=cov)

    for name, param in linear_layer.params.cov.named_parameters():
        if "bias" in name:
            assert param is None


@pytest.mark.parametrize(
    "parametrization",
    [
        params.Standard(),
        params.NeuralTangent(),
        params.MaximalUpdate(),
    ],
    ids=lambda x: x.__class__.__name__,
)
@pytest.mark.parametrize("layer_type", ["input", "hidden", "output"], ids=lambda x: x)
def test_weight_initialization(parametrization, layer_type):
    """Test whether the initialization of the weights is correct."""
    torch.manual_seed(2452345)

    with torch.no_grad():
        linear_layer = bnn.Conv3d(
            300,
            300,
            kernel_size=2,
            layer_type=layer_type,
            parametrization=parametrization,
            cov=None,
        )

        fan_in, fan_out = nn.init._calculate_fan_in_and_fan_out(
            linear_layer.params.weight
        )

        npt.assert_allclose(
            linear_layer.weight.mean(),
            0.0,
            atol=1e-2,
            rtol=1e-2,
        )

        npt.assert_allclose(
            linear_layer.weight.std(),
            parametrization.weight_init_scale(
                fan_in=fan_in, fan_out=fan_out, layer_type=layer_type
            ),
            atol=1e-2,
            rtol=1e-2,
        )


@pytest.mark.parametrize(
    "parametrization",
    [
        params.Standard(),
        params.NeuralTangent(),
        params.MaximalUpdate(),
    ],
    ids=lambda x: x.__class__.__name__,
)
@pytest.mark.parametrize("layer_type", ["input", "hidden", "output"], ids=lambda x: x)
def test_bias_initialization(parametrization, layer_type):
    """Test whether the initialization of the bias is correct."""
    torch.manual_seed(2452345)

    with torch.no_grad():
        linear_layer = bnn.Conv3d(
            300,
            300,
            kernel_size=2,
            bias=True,
            layer_type=layer_type,
            parametrization=parametrization,
            cov=None,
        )

        fan_in, fan_out = nn.init._calculate_fan_in_and_fan_out(
            linear_layer.params.weight
        )

        npt.assert_allclose(
            linear_layer.bias.mean(),
            0.0,
            atol=1e-1,
            rtol=1e-1,
        )

        npt.assert_allclose(
            linear_layer.bias.std(),
            parametrization.bias_init_scale(
                fan_in=fan_in, fan_out=fan_out, layer_type=layer_type
            ),
            atol=1e-1,
            rtol=1e-1,
        )


@pytest.mark.parametrize(
    "parametrization",
    [
        params.Standard(),
        params.NeuralTangent(),
        params.MaximalUpdate(),
    ],
    ids=lambda x: x.__class__.__name__,
)
@pytest.mark.parametrize("layer_type", ["input", "hidden", "output"], ids=lambda x: x)
@pytest.mark.parametrize(
    "layer_width,cov",
    [
        (10, params.FactorizedCovariance()),
        (100, params.LowRankCovariance(5)),
        (100, params.KroneckerCovariance()),
        (100, params.DiagonalCovariance()),
    ],
    ids=lambda x: (
        x.__class__.__name__
        if isinstance(x, params.FactorizedCovariance)
        else f"width_{x}"
    ),
)
def test_covariance_parameter_initialization(
    parametrization, layer_type, layer_width, cov
):
    """Test whether the initialization of the covariance parameters is correct."""
    torch.manual_seed(2452345)

    with torch.no_grad():
        linear_layer = bnn.Conv3d(
            layer_width,
            layer_width,
            kernel_size=2,
            bias=True,
            layer_type=layer_type,
            parametrization=parametrization,
            cov=cov,
        )

        fan_in, fan_out = nn.init._calculate_fan_in_and_fan_out(
            linear_layer.params.weight
        )

        if isinstance(cov, params.DiagonalCovariance):
            covariance_specific_scaling_factor = 1.0
        elif isinstance(cov, params.KroneckerCovariance):
            covariance_specific_scaling_factor = 1 / math.sqrt(cov.input_rank)
        elif isinstance(cov, params.LowRankCovariance):
            covariance_specific_scaling_factor = 1 / math.sqrt(cov.rank)
        elif isinstance(cov, params.FactorizedCovariance):
            covariance_specific_scaling_factor = 1 / math.sqrt(cov.rank)
        else:
            raise NotImplementedError()

        for name, param in linear_layer.params.cov.named_parameters():
            if param is not None:
                npt.assert_allclose(
                    param.mean(),
                    0.0,
                    atol=1e-1,
                    rtol=1e-1,
                )
            if "weight" in name:
                npt.assert_allclose(
                    param.std(),
                    parametrization.weight_init_scale(
                        fan_in=fan_in, fan_out=fan_out, layer_type=layer_type
                    )
                    * covariance_specific_scaling_factor,
                    atol=1e-1,
                    rtol=1e-1,
                )
            elif "bias" in name:
                npt.assert_allclose(
                    param.std(),
                    parametrization.bias_init_scale(
                        fan_in=fan_in, fan_out=fan_out, layer_type=layer_type
                    )
                    * covariance_specific_scaling_factor,
                    atol=1e-1,
                    rtol=1e-1,
                )



from __future__ import annotations

from .parametrization import LayerType, Parametrization, ParamType


class NeuralTangent(Parametrization):
    r"""Neural Tangent Parametrization (NTP).

    The neural tangent parametrization enables the theoretical analysis of the training dynamics of
    infinite-width neural networks (`Jacot et al., 2018`_) via the neural tangent kernel. However,
    NTP does not admit feature learning, as features are effectively fixed at initialization.

    .. _Jacot et al., 2018: http://arxiv.org/abs/1806.07572
    """

    def weight_init_scale(
        self,
        fan_in: int,
        fan_out: int,
        layer_type: LayerType = "hidden",
    ) -> float:
        return 1.0

    def bias_init_scale(
        self,
        fan_in: int,
        fan_out: int,
        layer_type: LayerType = "hidden",
    ) -> float:
        return 1.0

    def weight_lr_scale(
        self,
        fan_in: int,
        fan_out: int,
        optimizer: str,
        layer_type: LayerType = "hidden",
    ) -> float:
        # TODO: double check this scaling
        if optimizer == "SGD":
            if layer_type == "input":
                return 1.0
            elif layer_type == "output":
                return 1.0 / fan_in
            else:
                return 1.0 / fan_in
        elif optimizer == "Adam":
            raise NotImplementedError(
                f"LR scaling not implemented for optimizer '{optimizer}'."
            )
        else:
            raise NotImplementedError(
                f"LR scaling not implemented for optimizer '{optimizer}'."
            )

    def bias_lr_scale(
        self,
        fan_in: int,
        fan_out: int,
        optimizer: str,
        layer_type: LayerType = "hidden",
    ) -> float:
        # TODO: double check this scaling
        return 1 / fan_in



from numpy import testing as npt
import torch

from inferno.bnn import params

import pytest


@pytest.mark.parametrize(
    "mean_parameters",
    [
        {"weight": torch.randn(4, 2), "bias": torch.randn(4)},
        {"weight": torch.randn(1, 2), "bias": torch.randn((1,))},
        {"weight": torch.randn(1, 2, 3, 2), "bias": torch.randn((1, 3, 2))},
        {
            "weight": torch.randn(
                1,
            )
        },
    ],
)
def test_factor_matmul_low_rank_covariance(mean_parameters):
    """Test whether the multiplication with the covariance factor is correct."""
    torch.manual_seed(2345)

    covariance = params.DiagonalCovariance()
    numel_mean_parameters = sum(
        [tens.numel() for tens in mean_parameters.values() if tens is not None]
    )

    covariance.initialize_parameters(mean_parameters)
    covariance.reset_parameters({name: 1.0 for name in mean_parameters.keys()})

    input = torch.randn((3, 2, 1, numel_mean_parameters))
    matmul_result_dict = covariance.factor_matmul(
        input,
        additive_constant=torch.zeros(numel_mean_parameters),
    )

    stacked_scale_params = torch.hstack(
        [
            scale_param.view(-1)
            for scale_param in covariance.scale.values()
            if scale_param is not None
        ]
    )

    npt.assert_allclose(
        input * stacked_scale_params.detach().numpy(),
        torch.concatenate(
            [
                scale_param.view(*input.shape[:-1], -1)
                for scale_param in matmul_result_dict.values()
            ],
            dim=-1,
        )
        .detach()
        .numpy(),
    )



from __future__ import annotations

from typing import TYPE_CHECKING

import torch
from torch import nn

from inferno import bnn

if TYPE_CHECKING:
    from jaxtyping import Float
    from torch import Tensor


def _kl_normal_normal(
    loc_p: Float[Tensor, "parameter"],
    scale_p: Float[Tensor, "parameter parameter"],
    loc_q: Float[Tensor, "parameter"],
    scale_q: Float[Tensor, "parameter parameter"],
    eps: float = 1e-6,
):
    """KL Divergence between two normal distributions.

    :param loc_p:      Location of the first distribution.
    :param scale_p:    Scale of the first distribution.
    :param loc_q:      Location of the second distribution.
    :param scale_q:    Scale of the second distribution.
    :param eps:        Small value to avoid division by zero.
    """
    # Ensure that the variance is bounded away from zero
    # to avoid numerical instability
    var_p = scale_p.pow(2) + eps
    var_q = scale_q.pow(2) + eps
    # KL divergence between two normal distributions
    var_ratio = var_p / var_q
    quad_form = (loc_p - loc_q).pow(2) / var_q
    return 0.5 * (var_ratio + quad_form - 1.0 - var_ratio.log())


class VariationalFreeEnergy(nn.Module):
    """Variational Free Energy Loss.

    Computes the variational free energy loss for variational inference with the Kullback-Leibler
    regularization term computed in weight space. This is also known as the negative evidence lower
    bound (ELBO).

    :param nll:                 Loss function defining the negative log-likelihood.
    :param model:               The probabilistic model.
    :param prior_loc:           Location(s) of the prior Gaussian distribution.
    :param prior_scale:         Scale(s) of the prior Gaussian distribution.
    :param kl_weight:           Weight for the KL divergence term. If `None`, chooses the
        weight inversely proportional to the number of mean parameters.
    :param reduction:           Specifies the reduction to apply to the output:
        ````'mean'`` | ``'sum'``. ``'mean'``: the weighted mean of the output is taken,
        ``'sum'``: the output will be summed.
    """

    def __init__(
        self,
        nll: nn.modules.loss._Loss,
        model: bnn.BNNModule,
        prior_loc: Float[Tensor, "parameter"] | None = None,
        prior_scale: (
            Float[Tensor, "parameter"] | None
        ) = None,  # NOTE: Assumes mean-field prior for now.
        kl_weight: float | None = 1.0,
        reduction: str = "mean",
    ) -> None:
        super().__init__()
        self.nll = nll
        self.model = model
        self.numel_mean_parameters = sum(
            param.numel()
            for name, param in self.model.named_parameters()
            if param.requires_grad and "params." in name and "cov." not in name
        )
        self.prior_loc = prior_loc
        self.prior_scale = prior_scale
        if kl_weight is None:
            kl_weight = 1 / self.numel_mean_parameters
        self.kl_weight = kl_weight
        if reduction == "sum":
            raise NotImplementedError()
        self.nll.reduction = reduction
        self.reduction = reduction

    def forward(
        self,
        input: Float[Tensor, "*sample batch in_feature"],
        target: Float[Tensor, "batch out_feature"],
    ) -> Float[Tensor, ""]:
        # Expected negative log-likelihood (NLL)
        expected_nll = self.nll(
            input,
            target.expand((*input.shape[:-2], len(target))).reshape(-1),
        )  # NOTE: uses whatever reduction is set in the constructor

        # KL divergence regularization term
        variational_dist_mean = torch.concat(
            [
                param.ravel()
                for name, param in self.model.named_parameters()
                if param.requires_grad and "params." in name and "cov." not in name
            ]
        )
        variational_dist_scale = torch.concat(
            [
                param.ravel()
                for name, param in self.model.named_parameters()
                if param.requires_grad and "params.cov." in name
            ]
        )
        kl_divergence = _kl_normal_normal(
            loc_p=variational_dist_mean,
            scale_p=variational_dist_scale,
            loc_q=(
                self.prior_loc.to(device=variational_dist_mean.device)
                if self.prior_loc is not None
                else torch.zeros_like(variational_dist_mean)
            ),
            scale_q=(
                self.prior_scale.to(device=variational_dist_scale.device)
                if self.prior_scale is not None
                else torch.ones_like(variational_dist_scale)
            ),
        ).sum()
        if self.reduction == "mean":
            pass
        else:
            raise ValueError(f"Reduction '{self.reduction}' is not supported.")

        # Variational free energy (negative ELBO)
        return expected_nll + self.kl_weight * kl_divergence



import numpy.testing as npt
import torch
from torch import nn
import torchvision

import inferno
from inferno.bnn import params

import pytest


@pytest.mark.parametrize(
    "inferno_resnet,torchvision_resnet",
    [
        (
            inferno.models.ResNet(
                block=inferno.models.resnet.BasicBlock,
                num_blocks_per_layer=[2, 2, 2, 2],
                out_size=1000,
                cov=None,
            ),
            torchvision.models.ResNet(
                block=torchvision.models.resnet.BasicBlock,
                layers=[2, 2, 2, 2],
                norm_layer=lambda c: nn.GroupNorm(num_groups=32, num_channels=c),
            ),
        ),
        (
            inferno.models.ResNet(
                block=inferno.models.resnet.Bottleneck,
                num_blocks_per_layer=[2, 2, 2, 2],
                out_size=1000,
                cov=None,
            ),
            torchvision.models.ResNet(
                block=torchvision.models.resnet.Bottleneck,
                layers=[2, 2, 2, 2],
                norm_layer=lambda c: nn.GroupNorm(num_groups=32, num_channels=c),
            ),
        ),
    ],
)
def test_same_as_torchvision_resnet(inferno_resnet, torchvision_resnet):
    """Test whether the implementation matches the one of torchvision if no covariance is used."""
    torch.manual_seed(0)

    # Load weights from torchvision model
    state_dict = torchvision_resnet.state_dict()
    inferno_resnet.load_state_dict(state_dict, strict=True)

    # Create random input
    input = torch.randn((1, 3, 224, 224))

    # Forward pass through both models
    inferno_resnet.eval()
    torchvision_resnet.eval()
    with torch.no_grad():
        inferno_output = inferno_resnet(input, sample_shape=(2,))[
            0
        ]  # Draw multiple samples to check batch compatibility
        torchvision_output = torchvision_resnet(input)

    # Compare the outputs
    npt.assert_allclose(
        inferno_output.detach().numpy(),
        torchvision_output.detach().numpy(),
        rtol=1e-5,
        atol=1e-5,
    )


def test_draw_samples():
    """Test whether the model can draw samples."""
    torch.manual_seed(0)

    # Create a ResNet model
    model = inferno.models.ResNet(
        block=inferno.models.resnet.BasicBlock,
        num_blocks_per_layer=[2, 2, 2, 2],
        cov=params.KroneckerCovariance(),
        out_size=1000,
    )

    # Create random input
    batch_size = 8
    input = torch.randn((batch_size, 3, 224, 224))

    # Forward pass through the model
    model.eval()
    sample_shape = (10,)
    with torch.no_grad():
        output = model(input, sample_shape=sample_shape)

    # Check the shape of the output
    assert output.shape == (*sample_shape, batch_size, 1000)


def test_batch_norm_raises_value_error():
    """Test whether the model raises a ValueError when batch normalization is used with a non-batch dimension."""
    torch.manual_seed(0)

    with pytest.raises(ValueError):
        # Create a ResNet model with batch normalization
        model = inferno.models.ResNet(
            block=inferno.models.resnet.BasicBlock,
            num_blocks_per_layer=[2, 2, 2, 2],
            norm_layer=nn.BatchNorm2d,
            out_size=1000,
        )


@pytest.mark.parametrize(
    "resnet_type,weights,out_size,architecture,cov,freeze",
    [
        (
            inferno.models.ResNet18,
            torchvision.models.ResNet18_Weights.DEFAULT,
            10,
            "cifar",
            None,
            False,
        ),
        (
            inferno.models.ResNet18,
            torchvision.models.ResNet18_Weights.DEFAULT,
            100,
            "cifar",
            params.LowRankCovariance(100),
            True,
        ),
        (
            inferno.models.ResNet18,
            torchvision.models.ResNet18_Weights.DEFAULT,
            200,
            "imagenet",
            None,
            False,
        ),
        (
            inferno.models.ResNet34,
            torchvision.models.ResNet34_Weights.DEFAULT,
            1000,
            "imagenet",
            params.LowRankCovariance(100),
            True,
        ),
    ],
)
def test_from_pretrained_weights(
    resnet_type, weights, out_size, architecture, cov, freeze
):
    """Test whether the model can be loaded with pretrained weights."""
    torch.manual_seed(0)

    pretrained_model = resnet_type.from_pretrained_weights(
        weights=weights,
        out_size=out_size,
        architecture=architecture,
        cov=cov,
        freeze=freeze,
    )

    pretrained_weights_state_dict = weights.get_state_dict()

    # Check whether weights are loaded correctly
    npt.assert_allclose(
        pretrained_model.state_dict()["layer1.1.conv1.params.weight"].detach().numpy(),
        pretrained_weights_state_dict["layer1.1.conv1.weight"].detach().numpy(),
        rtol=1e-5,
        atol=1e-5,
    )

    npt.assert_allclose(
        pretrained_model.state_dict()["layer2.1.conv1.params.weight"].detach().numpy(),
        pretrained_weights_state_dict["layer2.1.conv1.weight"].detach().numpy(),
        rtol=1e-5,
        atol=1e-5,
    )

    # Check if freezing the weights works
    if freeze:
        for name, param in pretrained_model.named_parameters():
            if name.replace(".params", "") in pretrained_weights_state_dict:
                if name in [
                    "conv1.params.weight",
                    "fc.params.weight",
                    "fc.params.bias",
                ]:
                    # First and last layer may be trainable
                    continue

                assert not param.requires_grad, name



from __future__ import annotations

import copy
from typing import Callable

from torch import nn

from inferno import bnn
from inferno.bnn import params


class LeNet5(bnn.Sequential):
    """A simple convolutional neural network for image classification of 28x28 grayscale images.

    :param out_size:            Size of the output (i.e. number of classes).
    :param parametrization:     The parametrization to use. Defines the initialization
        and learning rate scaling for the parameters of the module.
    :param cov:                 Covariance structure of the weights.
    :param activation_layer:    Activation function following a linear layer.
    """

    def __init__(
        self,
        out_size: int = 10,
        parametrization: params.Parametrization = params.MaximalUpdate(),
        cov: params.FactorizedCovariance | None = None,
        activation_layer: Callable[..., nn.Module] | None = nn.ReLU,
    ) -> None:

        self.out_size = out_size

        # Layers
        layers = [
            bnn.Conv2d(
                1,
                6,
                kernel_size=5,
                padding=2,
                cov=(
                    copy.deepcopy(cov)
                    if isinstance(cov, params.DiagonalCovariance)
                    else copy.deepcopy(cov)
                ),
                layer_type="input",
            ),
            activation_layer() if activation_layer is not None else nn.Identity(),
            nn.MaxPool2d(kernel_size=2),
            bnn.Conv2d(
                6,
                16,
                kernel_size=5,
                cov=(
                    copy.deepcopy(cov)
                    if isinstance(cov, params.DiagonalCovariance)
                    else None
                ),
            ),
            activation_layer() if activation_layer is not None else nn.Identity(),
            nn.MaxPool2d(kernel_size=2),
            nn.Flatten(),
            bnn.Linear(
                16 * 5 * 5,
                120,
                cov=(
                    copy.deepcopy(cov)
                    if isinstance(cov, params.DiagonalCovariance)
                    else None
                ),
            ),
            activation_layer() if activation_layer is not None else nn.Identity(),
            bnn.Linear(
                120,
                84,
                cov=(
                    copy.deepcopy(cov)
                    if isinstance(cov, params.DiagonalCovariance)
                    else None
                ),
            ),
            activation_layer() if activation_layer is not None else nn.Identity(),
            bnn.Linear(
                84,
                out_size,
                cov=(
                    copy.deepcopy(cov)
                    if isinstance(cov, params.DiagonalCovariance)
                    else copy.deepcopy(cov)
                ),
                layer_type="output",
            ),
        ]

        super().__init__(*layers, parametrization=parametrization)



from __future__ import annotations

import os
from pathlib import Path
from typing import Any, Callable

import torchvision
from torchvision.datasets import VisionDataset


class TinyImageNet(VisionDataset):
    """TinyImageNet image classification dataset.

    The training dataset contains 100,000 images of 200 classes (500 for each class) downsized to 64x64 color images.
    The test set has 10,000 images (50 for each class).

    :param root: Root directory of the dataset.
    :param train: If True, creates dataset from training data, otherwise from test data.
    :param transform: Transform to apply to the data.
    :param target_transform: Transform to apply to the targets.
    :param download: If true, downloads the dataset from the internet and puts it in the root directory.
    """

    base_folder = Path("TinyImageNet/raw")
    tgz_md5 = "90528d7ca1a48142e341f4ef8d21d0de"
    url = "http://cs231n.stanford.edu/tiny-imagenet-200.zip"
    filename = "tiny-imagenet-200.zip"

    def __init__(
        self,
        root: str | Path,
        train: bool = True,
        transform: Callable | None = None,
        target_transform: Callable | None = None,
        download: bool = False,
    ) -> None:
        root = Path(root)

        super().__init__(
            root / self.base_folder,
            transform=transform,
            target_transform=target_transform,
        )

        if download:
            self.download()

        if not self._check_integrity():
            raise RuntimeError(
                f"Dataset '{self.__class__.__name__}' not found or corrupted. You can use download=True to download it."
            )

        self._image_root = root / self.base_folder / Path("tiny-imagenet-200")
        self._split = "train" if train else "val"

        # Construct the dataset
        with open(self._image_root / "wnids.txt") as r:
            classes = list(map(lambda s: s.strip(), r.readlines()))

        classes.sort()
        self.class_to_idx = {classes[i]: i for i in range(len(classes))}

        self.data = []
        data_dir = root / self.base_folder / Path("tiny-imagenet-200") / self._split
        if self._split == "train":
            for fname in data_dir.iterdir():
                if fname.is_dir():
                    image_folder = fname / Path("images")
                    for img_path in image_folder.iterdir():
                        self.data.append((img_path, self.class_to_idx[fname.stem]))
        else:
            imgs_path = os.path.join(data_dir, "images")
            imgs_annotations = os.path.join(data_dir, "val_annotations.txt")

            with open(imgs_annotations) as r:
                data_info = map(lambda s: s.split("\t"), r.readlines())

            cls_map = {line_data[0]: line_data[1] for line_data in data_info}

            for imgname in sorted(os.listdir(imgs_path)):
                path = os.path.join(imgs_path, imgname)
                item = (path, self.class_to_idx[cls_map[imgname]])
                self.data.append(item)

        self.targets = [item[1] for item in self.data]

    def _check_integrity(self) -> bool:
        """Check the integrity of the dataset."""
        return torchvision.datasets.utils.check_integrity(
            Path(self.root) / Path(self.filename), self.tgz_md5
        )

    def download(self) -> None:
        if self._check_integrity():
            # Check if files are already downloaded and not corrupted
            return
        else:
            # Download train and test set
            torchvision.datasets.utils.download_and_extract_archive(
                self.url, self.root, filename=self.filename, md5=self.tgz_md5
            )

    def __getitem__(self, index: int) -> tuple[Any, Any]:
        img_path, target = self.data[index]
        image = torchvision.datasets.folder.default_loader(img_path)

        if self.transform is not None:
            image = self.transform(image)
        if self.target_transform is not None:
            target = self.target_transform(target)

        return image, target

    def __len__(self):
        return len(self.data)


class TinyImageNetC(VisionDataset):
    """Corrupted TinyImageNet image classification dataset.

    Contains 10,000 64x64 color test images for each corruption (200 classes, 50 images per class).
    From [Benchmarking Neural Network Robustness to Common Corruptions and Perturbations](https://arxiv.org/abs/1903.12261).

    :param root: Root directory of the dataset.
    :param transform: Transform to apply to the data.
    :param target_transform: Transform to apply to the targets.
    :param corruptions: List of corruptions to apply to the data.
    :param shift_severity: Severity of the corruption to apply.
        Must be an integer between 1 and 5.
    :param download: If true, downloads the dataset from the internet and puts it in the root directory.
    """

    base_folder = Path("TinyImageNet-C/raw")
    tgz_md5 = [
        "f9c9a9dbdc11469f0b850190f7ad8be1",
        "0db0588d243cf403ef93449ec52b70eb",
    ]
    url = "https://zenodo.org/record/8206060/files/"
    filename = ["Tiny-ImageNet-C.tar", "Tiny-ImageNet-C-extra.tar"]

    def __init__(
        self,
        root: Path | str = None,
        transform: Callable | None = None,
        target_transform: Callable | None = None,
        corruptions: list[str] = [
            "brightness",
            "contrast",
            "defocus_blur",
            "elastic_transform",
            "fog",
            "frost",
            "gaussian_blur",
            "gaussian_noise",
            "glass_blur",
            "impulse_noise",
            "jpeg_compression",
            "motion_blur",
            "pixelate",
            "saturate",
            "shot_noise",
            "snow",
            "spatter",
            "speckle_noise",
            "zoom_blur",
        ],
        shift_severity: int = 5,
        download: bool = False,
    ):
        super().__init__(
            Path(root) / self.base_folder,
            transform=transform,
            target_transform=target_transform,
        )

        self.corruptions = corruptions
        if shift_severity not in list(range(1, 6)):
            raise ValueError(
                "Corruptions 'shift_severity' must be an integer between 1 and 5."
            )
        self.shift_severity = shift_severity

        if download:
            self.download()

        if not self._check_integrity():
            raise RuntimeError(
                f"Dataset '{self.__class__.__name__}' not found or corrupted. You can use download=True to download it."
            )

        # Construct the corrupted dataset
        self.data = []
        self.targets = []
        tiny_image_net = TinyImageNet(root, download=download)
        for corruption in self.corruptions:
            data_dir = (
                root
                / self.base_folder
                / Path(corruption)
                / Path(str(self.shift_severity))
            )
            for filename in data_dir.iterdir():
                if filename.is_dir():
                    for img_path in filename.iterdir():
                        self.data.append(img_path)
                        self.targets.append(tiny_image_net.class_to_idx[filename.stem])

    def _check_integrity(self) -> bool:
        """Check the integrity of the dataset."""
        for filename, md5 in list(zip(self.filename, self.tgz_md5, strict=True)):
            fpath = self.root / filename
            if not torchvision.datasets.utils.check_integrity(fpath, md5):
                return False
        return True

    def download(self) -> None:
        """Download the dataset."""
        if self._check_integrity():
            return
        for filename, md5 in list(zip(self.filename, self.tgz_md5, strict=True)):
            torchvision.datasets.utils.download_and_extract_archive(
                self.url + filename,
                self.root,
                filename=filename,
                md5=md5,
            )

        # Remove unnecessary subfolder
        for filename in (self.root / Path("Tiny-ImageNet-C")).iterdir():
            path = Path(filename).absolute()
            parent_dir = path.parents[1]
            path.rename(parent_dir / path.name)
        (self.root / Path("Tiny-ImageNet-C")).rmdir()

    def __getitem__(self, index: int) -> tuple[Any, Any]:
        img_path, target = self.data[index], self.targets[index]
        image = torchvision.datasets.folder.default_loader(img_path)

        if self.transform is not None:
            image = self.transform(image)
        if self.target_transform is not None:
            target = self.target_transform(target)

        return image, target

    def __len__(self) -> int:
        return len(self.data)



from __future__ import annotations

from pathlib import Path
from typing import Any, Callable

from PIL import Image
import numpy as np
import torchvision
from torchvision.datasets import CIFAR10, CIFAR100, VisionDataset


class CIFAR10C(VisionDataset):
    """Corrupted CIFAR10 image classification dataset.

    Contains 10,000 test images for each corruption.
    From [Benchmarking Neural Network Robustness to Common Corruptions and Perturbations](https://arxiv.org/abs/1903.12261).

    :param root: Root directory of the dataset.
    :param transform: Transform to apply to the data.
    :param target_transform: Transform to apply to the targets.
    :param corruptions: List of corruptions to apply to the data.
    :param shift_severity: Severity of the corruption to apply.
        Must be an integer between 1 and 5.
    :param download: If true, downloads the dataset from the internet and puts it in the root directory.
    """

    base_folder = Path("CIFAR10-C/raw")
    sub_folder = Path("CIFAR-10-C")
    tgz_md5 = "56bf5dcef84df0e2308c6dcbcbbd8499"
    url = "https://zenodo.org/record/2535967/files/CIFAR-10-C.tar"
    filename = "CIFAR-10-C.tar"

    corruption_data_checksums = {
        "fog": "7b397314b5670f825465fbcd1f6e9ccd",
        "jpeg_compression": "2b9cc4c864e0193bb64db8d7728f8187",
        "zoom_blur": "6ea8e63f1c5cdee1517533840641641b",
        "speckle_noise": "ef00b87611792b00df09c0b0237a1e30",
        "glass_blur": "7361fb4019269e02dbf6925f083e8629",
        "spatter": "8a5a3903a7f8f65b59501a6093b4311e",
        "shot_noise": "3a7239bb118894f013d9bf1984be7f11",
        "defocus_blur": "7d1322666342a0702b1957e92f6254bc",
        "elastic_transform": "9421657c6cd452429cf6ce96cc412b5f",
        "gaussian_blur": "c33370155bc9b055fb4a89113d3c559d",
        "frost": "31f6ab3bce1d9934abfb0cc13656f141",
        "saturate": "1cfae0964219c5102abbb883e538cc56",
        "brightness": "0a81ef75e0b523c3383219c330a85d48",
        "snow": "bb238de8555123da9c282dea23bd6e55",
        "gaussian_noise": "ecaf8b9a2399ffeda7680934c33405fd",
        "motion_blur": "fffa5f852ff7ad299cfe8a7643f090f4",
        "contrast": "3c8262171c51307f916c30a3308235a8",
        "impulse_noise": "2090e01c83519ec51427e65116af6b1a",
        "labels": "c439b113295ed5254878798ffe28fd54",
        "pixelate": "0f14f7e2db14288304e1de10df16832f",
    }

    def __init__(
        self,
        root: Path | str = None,
        transform: Callable | None = None,
        target_transform: Callable | None = None,
        corruptions: list[str] = [
            "brightness",
            "contrast",
            "defocus_blur",
            "elastic_transform",
            "fog",
            "frost",
            "gaussian_blur",
            "gaussian_noise",
            "glass_blur",
            "impulse_noise",
            "jpeg_compression",
            "motion_blur",
            "pixelate",
            "saturate",
            "shot_noise",
            "snow",
            "spatter",
            "speckle_noise",
            "zoom_blur",
        ],
        shift_severity: int = 5,
        download: bool = False,
    ):
        super().__init__(
            Path(root) / self.base_folder,
            transform=transform,
            target_transform=target_transform,
        )

        self.corruptions = corruptions
        if shift_severity not in list(range(1, 6)):
            raise ValueError(
                "Corruptions 'shift_severity' must be an integer between 1 and 5."
            )
        self.shift_severity = shift_severity

        if download:
            self.download()

        if not self._check_integrity():
            raise RuntimeError(
                "Dataset not found or corrupted. You can use download=True to download it."
            )

        # Construct the corrupted dataset
        self.data = np.concatenate(
            [
                np.load(self.root / self.sub_folder / (corruption + ".npy"))[
                    (self.shift_severity - 1) * 10000 : self.shift_severity * 10000
                ]
                for corruption in self.corruptions
            ],
            axis=0,
        )
        self.targets = np.tile(
            np.load(self.root / self.sub_folder / Path("labels.npy"))[
                (self.shift_severity - 1) * 10000 : self.shift_severity * 10000
            ],
            len(self.corruptions),
        ).astype(np.int64)

    def _check_integrity(self) -> bool:
        """Check the integrity of the dataset."""
        for corruption in self.corruptions:
            fpath = self.root / self.sub_folder / (corruption + ".npy")
            if corruption not in self.corruptions:
                raise ValueError(f"Unknown corruption '{corruption}'.")
            if not torchvision.datasets.utils.check_integrity(
                fpath, self.corruption_data_checksums[corruption]
            ):
                return False

        return True

    def download(self) -> None:
        if self._check_integrity():
            # Check if files are already downloaded and not corrupted
            return
        else:
            # Download train and test set
            torchvision.datasets.utils.download_and_extract_archive(
                self.url, self.root, filename=self.filename, md5=self.tgz_md5
            )

    def __getitem__(self, index: int) -> tuple[Any, Any]:

        img, target = self.data[index], self.targets[index]

        # Convert to PIL Image (for consistency with other datasets)
        img = Image.fromarray(img)

        if self.transform is not None:
            img = self.transform(img)

        if self.target_transform is not None:
            target = self.target_transform(target)

        return img, target

    def __len__(self) -> int:
        return len(self.data)


class CIFAR100C(CIFAR10C):
    """Corrupted CIFAR100 image classification dataset.

    From [Benchmarking Neural Network Robustness to Common Corruptions and Perturbations](https://arxiv.org/abs/1903.12261).

    :param root: Root directory of the dataset.
    :param transform: Transform to apply to the data.
    :param target_transform: Transform to apply to the targets.
    :param corruptions: List of corruptions to apply to the data.
    :param shift_severity: Severity of the corruption to apply.
        Must be an integer between 1 and 5.
    :param download: If true, downloads the dataset from the internet and puts it in root directory.
    """

    base_folder = Path("CIFAR100-C/raw")
    sub_folder = Path("CIFAR-100-C")
    tgz_md5 = "11f0ed0f1191edbf9fa23466ae6021d3"
    corruption_data_checksums = {
        "fog": "4efc7ebd5e82b028bdbe13048e3ea564",
        "jpeg_compression": "c851b7f1324e1d2ffddeb76920576d11",
        "zoom_blur": "0204613400c034a81c4830d5df81cb82",
        "speckle_noise": "e3f215b1a0f9fd9fd6f0d1cf94a7ce99",
        "glass_blur": "0bf384f38e5ccbf8dd479d9059b913e1",
        "spatter": "12ccf41d62564d36e1f6a6ada5022728",
        "shot_noise": "b0a1fa6e1e465a747c1b204b1914048a",
        "defocus_blur": "d923e3d9c585a27f0956e2f2ad832564",
        "elastic_transform": "a0792bd6581f6810878be71acedfc65a",
        "gaussian_blur": "5204ba0d557839772ef5a4196a052c3e",
        "frost": "3a39c6823bdfaa0bf8b12fe7004b8117",
        "saturate": "c0697e9fdd646916a61e9c312c77bf6b",
        "brightness": "f22d7195aecd6abb541e27fca230c171",
        "snow": "0237be164583af146b7b144e73b43465",
        "gaussian_noise": "ecc4d366eac432bdf25c024086f5e97d",
        "motion_blur": "732a7e2e54152ff97c742d4c388c5516",
        "contrast": "322bb385f1d05154ee197ca16535f71e",
        "impulse_noise": "3b3c210ddfa0b5cb918ff4537a429fef",
        "labels": "bb4026e9ce52996b95f439544568cdb2",
        "pixelate": "96c00c60f144539e14cffb02ddbd0640",
    }
    url = "https://zenodo.org/record/3555552/files/CIFAR-100-C.tar"
    filename = "CIFAR-100-C.tar"



from __future__ import annotations

import math

from .parametrization import LayerType, Parametrization, ParamType


class Standard(Parametrization):
    r"""Standard Parametrization (SP).

    The default parametrization for neural networks in PyTorch. Also known as the 'fan_in'
    or 'LeCun' initialization. 'Kaiming' initialization is the same up to multiplicative
    constants.
    """

    def weight_init_scale(
        self,
        fan_in: int,
        fan_out: int,
        layer_type: LayerType = "hidden",
    ) -> float:
        return 1 / math.sqrt(fan_in)

    def bias_init_scale(
        self,
        fan_in: int,
        fan_out: int,
        layer_type: LayerType = "hidden",
    ) -> float:
        # NOTE:  In Table 3 of [Tensor Programs V](http://arxiv.org/abs/2203.03466)
        # the bias in standard parametrization has scaling 1 / sqrt(fan_in) as
        # does the weight, but in the paper the fan_in of the bias is 1, rather
        # than the fan_in of the weight. This differs from the default initialization
        # in PyTorch which is 1.0 / math.sqrt(fan_in).
        return 1.0

    def weight_lr_scale(
        self,
        fan_in: int,
        fan_out: int,
        optimizer: str,
        layer_type: LayerType = "hidden",
    ) -> float:
        return 1.0

    def bias_lr_scale(
        self,
        fan_in,
        fan_out,
        optimizer: str,
        layer_type: LayerType = "hidden",
    ):
        return 1.0



"""Loss functions."""

from .variational_free_energy import VariationalFreeEnergy

VariationalFreeEnergy.__module__ = "inferno.loss_fns"
NegativeELBO = VariationalFreeEnergy

__all__ = [
    "VariationalFreeEnergy",
    "NegativeELBO",
]



from numpy import testing as npt
import torch

from inferno.bnn import params

import pytest


@pytest.mark.parametrize(
    "sample_shape", [(), (1,), (2, 3)], ids=lambda x: f"sample_shape_{str(x)}"
)
@pytest.mark.parametrize(
    "covariance,mean_parameters",
    [
        (
            params.KroneckerCovariance(),
            {"weight": torch.randn(4, 2), "bias": torch.randn(4)},
        ),
        (
            params.KroneckerCovariance(input_rank=2, output_rank=3),
            {"weight": torch.randn(5, 2, 3, 2), "bias": torch.randn((5,))},
        ),
        (
            params.KroneckerCovariance(),
            {"weight": torch.randn(1, 2), "bias": torch.randn((1,))},
        ),
        (
            params.KroneckerCovariance(input_rank=99, output_rank=None),
            {"weight": torch.randn(1, 1)},
        ),
    ],
    ids=["linear", "conv2d", "linear_1d", "linear_large_rank"],
)
def test_factor_matmul(sample_shape, covariance, mean_parameters):
    """Test whether the multiplication with the covariance factor is correct."""

    torch.manual_seed(2435)

    numel_parameters = sum(
        [tens.numel() for tens in mean_parameters.values() if tens is not None]
    )

    covariance.initialize_parameters(mean_parameters)
    covariance.reset_parameters({name: 1.0 for name in mean_parameters.keys()})

    input = torch.randn(*sample_shape, covariance.rank)

    # Efficient Kronecker matmul
    matmul_result_dict = covariance.factor_matmul(
        input, additive_constant=torch.zeros(numel_parameters)
    )

    # Dense Kronecker matmul
    stacked_input_factor, stacked_output_factor = covariance._stacked_parameters()
    kronecker_product_of_factors = torch.kron(
        stacked_input_factor, stacked_output_factor
    )
    result_dense_kron = (kronecker_product_of_factors @ input.unsqueeze(-1)).squeeze(-1)

    dense_matmul_result_dict = {}
    starting_idx_weight_result = 0
    if "bias" in mean_parameters.keys():
        dense_matmul_result_dict["bias"] = result_dense_kron[
            ..., : covariance.output_factor.shape[0]
        ]
        starting_idx_weight_result = covariance.output_factor.shape[0]
    if "weight" in mean_parameters.keys():
        dense_matmul_result_dict["weight"] = result_dense_kron[
            ..., starting_idx_weight_result:
        ].view(
            *sample_shape,
            covariance.output_factor.shape[0],
            *mean_parameters["weight"].shape[1:],
        )

    for name, tens in matmul_result_dict.items():
        npt.assert_allclose(
            tens.detach(), dense_matmul_result_dict[name].detach(), atol=1e-5, rtol=1e-5
        )



import copy

import numpy.testing as npt
import torch
from torch import nn

from inferno import bnn, models
from inferno.bnn import params

import pytest


@pytest.mark.parametrize(
    "base_module",
    [
        nn.Linear(5, 3),
        bnn.Linear(5, 6),
        bnn.Sequential(
            bnn.Linear(5, 3, cov=params.FactorizedCovariance()),
            nn.ReLU(),
            nn.Linear(3, 2),
            nn.Softmax(dim=-1),
        ),
    ],
)
@pytest.mark.parametrize("ensemble_size", [1, 3, 10])
@pytest.mark.parametrize("sample_shape", [(), (1,), (3, 2)])
@pytest.mark.parametrize("batch_shape", [(1,), (4,)])
def test_shape(base_module, ensemble_size, sample_shape, batch_shape):
    """Test whether the output shape of the ensemble is correct."""
    generator = torch.Generator().manual_seed(0)

    ensemble = models.Ensemble(
        members=[copy.deepcopy(base_module) for _ in range(ensemble_size)],
    )
    if isinstance(base_module, bnn.Sequential):
        in_features = base_module[0].in_features
    elif isinstance(base_module, (bnn.Linear, nn.Linear)):
        in_features = base_module.in_features
    else:
        raise ValueError()

    input = torch.randn((*batch_shape, in_features), generator=generator)
    output = ensemble(input, sample_shape=sample_shape, generator=generator)

    if isinstance(base_module, bnn.BNNModule):
        kwargs = {"generator": generator}
    else:
        kwargs = {}
        sample_shape = ()

    assert output.shape == (
        ensemble_size,
        *sample_shape,
        *batch_shape,
        base_module(torch.randn((in_features,), generator=generator), **kwargs).shape[
            -1
        ],
    )


@pytest.mark.parametrize(
    "base_module",
    [
        nn.Linear(4, 1),
        bnn.Linear(4, 1, cov=params.FactorizedCovariance()),
        bnn.Sequential(
            bnn.Linear(4, 3, cov=params.LowRankCovariance(2)),
            nn.ReLU(),
            bnn.Linear(3, 1, cov=params.FactorizedCovariance()),
        ),
    ],
)
@pytest.mark.parametrize("seed", [0, 45234, 42])
def test_forward_is_deterministic_given_generator(base_module, seed):
    """Test whether the forward method is deterministic given a generator."""
    ensemble = models.Ensemble(
        members=[copy.deepcopy(base_module) for _ in range(10)],
    )

    input = torch.randn(3, 4, generator=torch.Generator().manual_seed(seed + 2452345))
    output1 = ensemble(input, generator=torch.Generator().manual_seed(seed))
    output2 = ensemble(input, generator=torch.Generator().manual_seed(seed))

    npt.assert_allclose(output1.detach().numpy(), output2.detach().numpy())


@pytest.mark.parametrize(
    "members,parametrization",
    [
        (
            [
                bnn.Linear(
                    5,
                    3,
                    cov=params.FactorizedCovariance(),
                    parametrization=params.Standard(),
                )
            ]
            * 5,
            params.Standard,
        ),
        (
            [
                bnn.Linear(
                    5,
                    3,
                    cov=params.FactorizedCovariance(),
                    parametrization=params.MaximalUpdate(),
                )
            ]
            * 5,
            params.MaximalUpdate,
        ),
        ([nn.Linear(5, 3)] * 5, params.Standard),
    ],
)
def test_parametrization(members, parametrization):
    """Test whether the parametrization of the ensemble is correct."""
    ensemble = models.Ensemble(members)

    assert isinstance(ensemble.parametrization, parametrization)

    for member in ensemble.members:
        if hasattr(member, "parametrization"):
            # Check that the parametrization of the member is the same as the ensemble
            # (if it has one)
            assert isinstance(member.parametrization, parametrization)



from __future__ import annotations

from .factorized import FactorizedCovariance


class LowRankCovariance(FactorizedCovariance):
    r"""Covariance of a Gaussian parameter with low-rank structure.

    Assumes the covariance is factorized as a product of a matrix
    :math:\mathbf{S} \in \mathbb{R}^{P \times R}` and its transpose.

    ..math::
        \mathbf{\Sigma} = \mathbf{S} \mathbf{S}^\top


    :param rank: Rank of the covariance matrix. If `None`, the rank is set to the total number
        of mean parameters.
    """

    def __init__(
        self,
        rank: int,
    ):
        super().__init__(rank=rank)



from numpy import testing as npt
import torch
from torch import nn

from inferno import bnn
from inferno.bnn import params

import pytest


@pytest.mark.parametrize(
    "model",
    [
        bnn.Sequential(
            bnn.Linear(5, 3, layer_type="input"),
            bnn.Linear(3, 2, cov=params.FactorizedCovariance()),
        ),
        bnn.Sequential(
            bnn.Linear(5, 3, cov=params.LowRankCovariance(2)),
            nn.ReLU(),
            nn.Linear(3, 2),
            nn.Softmax(dim=-1),
        ),
        bnn.Sequential(
            nn.Linear(5, 3),
        ),
        bnn.Sequential(
            nn.Linear(5, 6),
            nn.Flatten(0, 0),
            bnn.Linear(6, 1, cov=params.FactorizedCovariance(), layer_type="output"),
        ),
    ],
)
@pytest.mark.parametrize("sample_shape", [(), (1,), (2,), (3, 2)])
@pytest.mark.parametrize("batch_shape", [(1,), (3,)])
def test_shape(model, sample_shape, batch_shape):
    """Test whether the output shape of a container module is correct."""
    generator = torch.Generator().manual_seed(0)
    input = torch.randn(*batch_shape, model[0].in_features, generator=generator)
    output = model(input, sample_shape=sample_shape, generator=generator)
    assert output.shape == (
        *sample_shape,
        *batch_shape,
        model(torch.randn(model[0].in_features, generator=generator)).shape[-1],
    )


@pytest.mark.parametrize("seed", [0, 45234, 42])
def test_forward_is_deterministic_given_generator(seed):
    """Test whether the forward method is deterministic given a generator."""
    sequential_model = bnn.Sequential(
        bnn.Linear(4, 3),
        nn.ReLU(),
        bnn.Linear(3, 1, cov=params.FactorizedCovariance()),
    )

    input = torch.randn(3, 4, generator=torch.Generator().manual_seed(seed + 2452345))
    output1 = sequential_model(input, generator=torch.Generator().manual_seed(seed))
    output2 = sequential_model(input, generator=torch.Generator().manual_seed(seed))

    npt.assert_allclose(output1.detach().numpy(), output2.detach().numpy())


@pytest.mark.parametrize(
    "sequential_to_load,new_sequential,strict",
    [
        (
            nn.Sequential(
                nn.Linear(5, 3),
                nn.SiLU(),
                nn.Linear(3, 2),
            ),
            bnn.Sequential(
                bnn.Linear(5, 3, cov=params.FactorizedCovariance()),
                nn.SiLU(),
                nn.Linear(3, 2),
            ),
            False,
        ),
        (
            bnn.Sequential(
                bnn.Linear(5, 3, cov=params.FactorizedCovariance(), layer_type="input"),
                nn.ReLU(),
                bnn.Linear(
                    3, 2, cov=params.FactorizedCovariance(), layer_type="output"
                ),
                nn.Softmax(),
            ),
            bnn.Sequential(
                bnn.Linear(5, 3, cov=params.FactorizedCovariance(), layer_type="input"),
                nn.ReLU(),
                bnn.Linear(
                    3, 2, cov=params.FactorizedCovariance(), layer_type="output"
                ),
                nn.Softmax(),
            ),
            True,
        ),
    ],
)
def test_load_from_state_dict(sequential_to_load, new_sequential, strict):
    """Test whether the load_from_state_dict method is working for torch and inferno
    sequential layers."""
    torch.manual_seed(53124)

    missing_keys, unexpected_keys = new_sequential.load_state_dict(
        sequential_to_load.state_dict(),
        strict=strict,
    )

    assert len(unexpected_keys) == 0
    assert len(missing_keys) == 0 if strict else True


@pytest.mark.parametrize("layer_idx", [0, 1, 2, 3])
def test_register_forward_hook(layer_idx):
    """Test whether the register_forward_hook method is working on each layer of the
    sequential container."""
    model = bnn.Sequential(
        bnn.Linear(5, 3, cov=params.FactorizedCovariance()),
        nn.ReLU(),
        bnn.Linear(3, 2),
        nn.Softmax(dim=-1),
    )
    test_dict = {"hook_has_fired": False}

    def hook(module, input, output):
        test_dict["hook_has_fired"] = True

    model[layer_idx].register_forward_hook(hook)
    model(torch.randn(3, 5))

    assert test_dict["hook_has_fired"]


def test_setting_a_parametrization_overrides_module_parametrizations():
    """Test whether setting a parametrization on the sequential module overrides the
    parametrizations of the modules in the container."""
    model = bnn.Sequential(
        bnn.Linear(5, 3, cov=params.FactorizedCovariance()),
        nn.ReLU(),
        bnn.Linear(3, 2, parametrization=params.Standard()),
        parametrization=params.MaximalUpdate(),
    )

    for module in model.modules():
        if isinstance(module, bnn.BNNModule):
            assert isinstance(module.parametrization, params.MaximalUpdate)

    model.parametrization = params.NeuralTangent()

    for module in model.modules():
        if isinstance(module, bnn.BNNModule):
            assert isinstance(module.parametrization, params.NeuralTangent)


def test_no_parametrization_given():
    """Test whether modules retain their parametrization if no parametrization is given."""
    parametrizations = [
        params.Standard,
        params.MaximalUpdate,
    ]
    model = bnn.Sequential(
        bnn.Linear(
            5,
            3,
            cov=params.FactorizedCovariance(),
            parametrization=parametrizations[0](),
        ),
        bnn.Linear(3, 2, parametrization=parametrizations[1]()),
    )

    for i, module in enumerate(model):
        if isinstance(module, bnn.BNNModule):
            assert isinstance(module.parametrization, parametrizations[i])



from __future__ import annotations

import math
from typing import TYPE_CHECKING

import torch

from .parameter import BNNParameter

if TYPE_CHECKING:
    from jaxtyping import Float
    from torch import Tensor

    from .covariances.factorized import FactorizedCovariance


class GaussianParameter(BNNParameter):
    """Parameter of a BNNModule with Gaussian distribution.

    :param mean: Mean of the Gaussian distribution.
    :param cov: Covariance of the Gaussian distribution.
    """

    def __init__(
        self,
        mean: Float[Tensor, "parameter"] | dict[str, Float[Tensor, "parameter"]],
        cov: FactorizedCovariance,
    ):
        if not isinstance(mean, dict):
            mean = {"mean": mean}

        super().__init__(hyperparameters=mean)

        self.cov = cov
        self.cov.initialize_parameters(mean)

    def sample(
        self,
        sample_shape: torch.Size = torch.Size([]),
        generator: torch.Generator | None = None,
    ) -> (
        Float[Tensor, "*sample parameter"]
        | dict[str, Float[Tensor, "*sample parameter"]]
    ):
        # Sample from standard normal distribution
        standard_normal_sample = torch.randn(
            sample_shape + (self.cov.rank,),
            dtype=next(self.cov.parameters()).dtype,
            device=next(self.cov.parameters()).device,
            generator=generator,
        )

        # Transform the standard normal sample to the correct mean and covariance
        mean_params = {
            name: tens
            for name, tens in self.named_parameters()
            if "cov." not in name and "temperature" not in name
        }

        mean_params_stacked = torch.hstack(
            [tens.view(-1) for tens in mean_params.values()]
        )

        # Scale with inverse temperature if not training and the parameters are in the output layer
        if hasattr(self, "temperature") and not self.training:
            mean_params_stacked = mean_params_stacked / self.temperature

        return self.cov.factor_matmul(
            standard_normal_sample,
            additive_constant=mean_params_stacked,
        )



import copy

import numpy.testing as npt
import torch
from torch import nn

from inferno import bnn
from inferno.bnn import params

import pytest

model_loss_fn_input_dim_num_classes_tuples = [
    (
        bnn.Sequential(
            nn.Linear(10, 10),
            nn.ReLU(),
            bnn.Linear(
                10,
                2,
                layer_type="output",
            ),
        ),
        nn.CrossEntropyLoss(),
        10,
        2,
    ),
    (
        bnn.Sequential(
            bnn.Linear(
                10,
                10,
                cov=params.FactorizedCovariance(),
                layer_type="input",
            ),
            nn.ReLU(),
            bnn.Linear(
                10,
                3,
                cov=params.FactorizedCovariance(),
                layer_type="output",
            ),
        ),
        nn.CrossEntropyLoss(),
        10,
        3,
    ),
    (
        bnn.Sequential(
            bnn.Linear(
                10,
                10,
                cov=params.FactorizedCovariance(),
                layer_type="input",
            ),
            nn.ReLU(),
            bnn.Linear(
                10,
                1,
                cov=params.FactorizedCovariance(),
                layer_type="output",
            ),
            nn.Flatten(-2, -1),
        ),
        nn.BCEWithLogitsLoss(),
        10,
        2,
    ),
]


@pytest.mark.parametrize(
    "model,loss_fn,input_dim,num_classes",
    model_loss_fn_input_dim_num_classes_tuples,
    ids=["deterministic", "probabilistic", "bce_logits"],
)
def test_temperature_scaler_on_different_models(model, loss_fn, input_dim, num_classes):

    torch.manual_seed(0)

    # Create a temperature scaler
    temperature_scaler = bnn.TemperatureScaler(loss_fn=loss_fn)

    # Create a dummy dataloader
    num_data = 50
    dataloader = torch.utils.data.DataLoader(
        torch.utils.data.TensorDataset(
            torch.randn(num_data, input_dim),
            torch.randint(
                0,
                num_classes,
                (num_data,),
                dtype=(
                    torch.float32
                    if isinstance(loss_fn, nn.BCEWithLogitsLoss)
                    else torch.long
                ),
            ),
        ),
        batch_size=10,
    )

    temperature_param = None
    for name, param in model.named_parameters():
        if "temperature" in name:
            temperature_param = param
            break

    assert temperature_param.item() == 1.0

    # Optimize the temperature
    temperature_scaler.optimize(model, dataloader)

    assert temperature_param.item() != 1.0

    for name, param in model.named_parameters():
        if "temperature" in name:
            assert not param.requires_grad
        else:
            assert param.requires_grad


@pytest.mark.parametrize(
    "model,loss_fn,input_dim,num_classes",
    model_loss_fn_input_dim_num_classes_tuples,
    ids=["deterministic", "probabilistic", "bce_logits"],
)
def test_temperature_scaler_doesnt_modify_model_parameters(
    model, loss_fn, input_dim, num_classes
):
    torch.manual_seed(0)

    # Create a temperature scaler
    temperature_scaler = bnn.TemperatureScaler(loss_fn=loss_fn)

    # Create a dummy dataloader
    num_data = 50
    dataloader = torch.utils.data.DataLoader(
        torch.utils.data.TensorDataset(
            torch.randn(num_data, input_dim),
            torch.randint(
                0,
                num_classes,
                (num_data,),
                dtype=(
                    torch.float32
                    if isinstance(loss_fn, nn.BCEWithLogitsLoss)
                    else torch.long
                ),
            ),
        ),
        batch_size=10,
    )

    temperature_param = None
    model_params_copy = {}
    for name, param in model.named_parameters():
        if "temperature" in name:
            temperature_param = param
        model_params_copy[name] = copy.deepcopy(param)

    # Optimize the temperature
    temperature_scaler.optimize(model, dataloader)

    # Check that the model parameters are not modified
    for name, param in model.named_parameters():
        if "temperature" not in name:
            npt.assert_array_equal(
                model_params_copy[name].detach().cpu().numpy(),
                param.detach().cpu().numpy(),
            )


@pytest.mark.parametrize(
    "model,loss_fn,input_dim,num_classes",
    model_loss_fn_input_dim_num_classes_tuples,
    ids=["deterministic", "probabilistic", "bce_logits"],
)
@pytest.mark.parametrize("training_mode", [True, False])
def test_temperature_scaling_works_regardless_of_training_or_eval_mode(
    model, loss_fn, input_dim, num_classes, training_mode
):
    torch.manual_seed(0)

    # Create a temperature scaler
    temperature_scaler = bnn.TemperatureScaler(loss_fn=loss_fn)

    # Create a dummy dataloader
    num_data = 50
    dataloader = torch.utils.data.DataLoader(
        torch.utils.data.TensorDataset(
            torch.randn(num_data, input_dim),
            torch.randint(
                0,
                num_classes,
                (num_data,),
                dtype=(
                    torch.float32
                    if isinstance(loss_fn, nn.BCEWithLogitsLoss)
                    else torch.long
                ),
            ),
        ),
        batch_size=10,
    )

    # Set the model to training or evaluation mode
    if training_mode:
        model.train()
    else:
        model.eval()

    # Optimize the temperature
    temperature_scaler.optimize(model, dataloader)

    assert model.training == training_mode



from numpy import testing as npt
import torch

from inferno.bnn import params

import pytest


@pytest.mark.parametrize(
    "covariance,mean_parameters",
    [
        (
            params.FactorizedCovariance(),
            {"weight": torch.randn(4, 2), "bias": torch.randn(4)},
        ),
        (
            params.FactorizedCovariance(),
            {"weight": torch.randn(1, 2, 3, 2), "bias": torch.randn((1, 3, 2))},
        ),
        (
            params.FactorizedCovariance(),
            {"weight": torch.randn(1, 2), "bias": torch.randn((1,))},
        ),
        (
            params.FactorizedCovariance(),
            {
                "weight": torch.randn(
                    1,
                )
            },
        ),
    ],
)
def test_factor_matmul(covariance, mean_parameters):
    """Test whether the multiplication with the covariance factor is correct."""

    torch.manual_seed(2435)

    numel_parameters = sum(
        [tens.numel() for tens in mean_parameters.values() if tens is not None]
    )

    covariance.initialize_parameters(mean_parameters)
    covariance.reset_parameters({name: 1.0 for name in mean_parameters.keys()})

    matmul_result_dict = covariance.factor_matmul(
        torch.eye(covariance.rank), additive_constant=torch.zeros(numel_parameters)
    )

    for name, tens in matmul_result_dict.items():
        npt.assert_allclose(
            torch.movedim(tens, 0, -1)
            .detach()
            .numpy(),  # Rank dimension is in front for factor_matmul
            covariance.factor[name].detach().numpy(),
        )



from __future__ import annotations

from pathlib import Path
from typing import Any, Callable

from PIL import Image
import numpy as np
import torchvision
from torchvision.datasets import MNIST, VisionDataset


class MNISTC(VisionDataset):
    """Corrupted MNIST image classification dataset.

    Contains 10,000 test images for each one of 15 corruptions.
    From [MNIST-C: A Robustness Benchmark for Computer Vision](https://arxiv.org/abs/1906.02337).

    :param root: Root directory of the dataset.
    :param transform: Transform to apply to the data.
    :param target_transform: Transform to apply to the targets.
    :param corruptions: List of corruptions to apply to the data.
    :param download: If true, downloads the dataset from the internet and puts it in the root directory.
    """

    base_folder = Path("MNIST-C/raw")
    sub_folder = Path("mnist_c")
    zip_md5 = "4b34b33045869ee6d424616cd3a65da3"
    url = "https://zenodo.org/record/3239543/files/mnist_c.zip"
    filename = "mnist_c.zip"

    def __init__(
        self,
        root: Path | str = None,
        transform: Callable | None = None,
        target_transform: Callable | None = None,
        corruptions: list[str] = [
            "brightness",
            "canny_edges",
            "dotted_line",
            "fog",
            "glass_blur",
            "impulse_noise",
            "motion_blur",
            "rotate",
            "scale",
            "shear",
            "shot_noise",
            "spatter",
            "stripe",
            "translate",
            "zigzag",
        ],
        download: bool = False,
    ):
        super().__init__(
            Path(root) / self.base_folder,
            transform=transform,
            target_transform=target_transform,
        )

        self.corruptions = corruptions

        if download:
            self.download()

        if not self._check_integrity():
            raise RuntimeError(
                "Dataset not found or corrupted. You can use download=True to download it."
            )

        # Construct the corrupted dataset
        self.data = np.concatenate(
            [
                np.load(self.root / self.sub_folder / corruption / "test_images.npy")
                for corruption in self.corruptions
            ],
            axis=0,
        )
        self.targets = np.tile(
            np.load(self.root / self.sub_folder / "identity/test_labels.npy"),
            len(self.corruptions),
        ).astype(np.int64)

    def _check_integrity(self) -> bool:
        """Check the integrity of the dataset."""
        return torchvision.datasets.utils.check_integrity(
            self.root / self.filename, self.zip_md5
        )

    def download(self) -> None:
        if self._check_integrity():
            # Check if files are already downloaded and not corrupted
            return
        else:
            # Download train and test set
            torchvision.datasets.utils.download_and_extract_archive(
                self.url, self.root, filename=self.filename, md5=self.zip_md5
            )

    def __getitem__(self, index: int) -> tuple[Any, Any]:

        img, target = self.data[index], self.targets[index]

        # Convert to PIL Image (for consistency with other datasets)
        img = Image.fromarray(img[..., -1], mode="L")

        if self.transform is not None:
            img = self.transform(img)

        if self.target_transform is not None:
            target = self.target_transform(target)

        return img, target

    def __len__(self) -> int:
        return len(self.data)



from __future__ import annotations

import math

from .parametrization import LayerType, Parametrization, ParamType


class MaximalUpdate(Parametrization):
    r"""Maximal update parametrization ($\mu P$).

    The maximal update parametrization (`Yang et al., 2021`_, `Yang et al., 2021b`_) enforces:
        - *stable training dynamics*, meaning (pre)activations and logits at activations are independent
          of width at initialization and features and logits do not explode during training.
        - *feature learning*, meaning the model can learn representations from the data at any width.

    This parametrization is particularly useful when training large models, since it maintains
    stable training dynamics at any width and enables hyperparameter transfer. This means one can
    tune the learning rate on a smaller model and achieve good generalization with the tuned
    learning rate for a large model.

    .. _Yang et al., 2021: http://arxiv.org/abs/2011.14522
    .. _Yang et al., 2021b: http://arxiv.org/abs/2203.03466
    """

    def weight_init_scale(
        self,
        fan_in: int,
        fan_out: int,
        layer_type: LayerType = "hidden",
    ) -> float:
        if layer_type == "input":
            return 1.0 / math.sqrt(fan_in)
        elif layer_type == "output":
            return 1.0 / fan_in
        else:
            return 1.0 / math.sqrt(fan_in)

    def bias_init_scale(
        self,
        fan_in: int,
        fan_out: int,
        layer_type: LayerType = "hidden",
    ) -> float:
        # NOTE: In Table 3 of [Tensor Programs V](http://arxiv.org/abs/2203.03466),
        # the bias has scaling 1.0 / sqrt(fan_in) as does the weight, but in the paper the
        # fan_in of the bias is 1, rather than the fan_in of the weight.
        return 1.0

    def weight_lr_scale(
        self,
        fan_in: int,
        fan_out: int,
        optimizer: str,
        layer_type: LayerType = "hidden",
    ) -> float:

        if optimizer == "SGD":
            if layer_type == "input":
                return fan_out
            elif layer_type == "output":
                return 1.0 / fan_in
            else:
                return 1.0
        elif optimizer == "Adam":
            if layer_type == "input":
                return 1.0
            elif layer_type == "output":
                return 1.0 / fan_in
            else:
                return 1.0 / fan_in
        else:
            raise NotImplementedError(
                f"LR scaling not implemented for optimizer '{optimizer}'."
            )

    def bias_lr_scale(
        self,
        fan_in: int,
        fan_out: int,
        optimizer: str,
        layer_type: LayerType = "hidden",
    ) -> float:
        if optimizer == "SGD":
            return fan_out
        elif optimizer == "Adam":
            return 1.0
        else:
            raise NotImplementedError(
                f"LR scaling not implemented for optimizer '{optimizer}'."
            )



from __future__ import annotations

import math
from typing import TYPE_CHECKING

import numpy as np
import torch
from torch import nn

from .factorized import FactorizedCovariance

if TYPE_CHECKING:
    from jaxtyping import Float
    from torch import Tensor


class DiagonalCovariance(FactorizedCovariance):
    r"""Covariance of a Gaussian parameter with diagonal structure."""

    def __init__(self):
        super().__init__(rank=None)

    def initialize_parameters(
        self,
        mean_parameters: dict[str, torch.Tensor],
    ) -> None:
        if self.rank is None:
            # Total number of mean parameters
            self.rank = sum(
                [tens.numel() for tens in mean_parameters.values() if tens is not None]
            )

        # Initialize scale parameters
        self.scale = nn.ParameterDict(
            {
                name: (
                    torch.empty(
                        *param.shape,
                        dtype=param.dtype,
                        device=param.device,
                    )
                    if param is not None
                    else None
                )
                for name, param in mean_parameters.items()
            }
        )

    def reset_parameters(
        self,
        mean_parameter_scales: dict[str, float] | float = 1.0,
    ) -> None:
        if isinstance(mean_parameter_scales, float):
            mean_parameter_scales = {
                name: mean_parameter_scales for name in self.scale.keys()
            }

        for name, param in self.scale.items():
            if param is not None:
                nn.init.normal_(param, mean=0, std=mean_parameter_scales[name])

    def factor_matmul(
        self,
        input: Float[Tensor, "*sample parameter"],
        /,
        additive_constant: Float[Tensor, "*sample parameter"] | None = None,
    ) -> dict[str, Float[Tensor, "*sample parameter"]]:
        sample_shape = input.shape[:-1]

        # Stack scale parameters
        stacked_parameters = self._stacked_parameters()

        # Multiply scale parameters with input
        result = torch.einsum("...p,p->...p", input, stacked_parameters)
        if additive_constant is not None:
            result = result + additive_constant

        # Split result into parameter shapes
        split_result = torch.tensor_split(
            result,
            list(
                np.cumsum(
                    [tens.numel() for tens in self.scale.values() if tens is not None]
                )[:-1]
            ),
            dim=-1,
        )

        result_dict = {}
        i = 0
        for name, param in self.scale.items():
            if param is not None:
                result_dict[name] = split_result[i].view(*sample_shape, *param.shape)
                i += 1

        return result_dict

    def _stacked_parameters(self) -> Float[Tensor, "parameter"]:
        """Stack parameters into a single tensor."""
        stacked_parameters = torch.hstack(
            [
                scale_param.view(-1)
                for scale_param in self.scale.values()
                if scale_param is not None
            ]
        )
        return stacked_parameters

    def to_dense(self) -> Float[Tensor, "parameter parameter"]:
        """Convert the covariance matrix to a dense representation."""
        stacked_parameters = self._stacked_parameters()
        return torch.diag(stacked_parameters**2)

    @property
    def lr_scaling(self) -> dict[str, float]:
        return {"scale." + name: 1.0 for name in self.scale.keys()}
